# vim: ts=4:sw=4:expandtab
# -*- coding: UTF-8 -*-

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Preferences dialog
"""

# standard imports
import logging
import os

# first party imports
from bleachbit import GuiBasic
from bleachbit import online_update_notification_enabled
from bleachbit import ProtectedPath
from bleachbit.Constant import EMPTY_SPACE_WARNING, REQUIRES_EXPERT_MODE
from bleachbit.GtkShim import Gtk, GLib
from bleachbit.GuiCookie import CookieManagerPane
from bleachbit.GuiUtil import (detect_dark_background, flush_gtk_events,
                               should_show_dark_mode_warning)
from bleachbit.Language import get_active_language_code, get_supported_language_code_name_dict, setup_translation
from bleachbit.Language import get_text as _, pget_text as _p
from bleachbit.Options import options

logger = logging.getLogger(__name__)

LOCATIONS_WHITELIST = 1
LOCATIONS_CUSTOM = 2
# TRANSLATORS: Shown as a tooltip in the preferences window and as
# a description in a confirmation dialog.
EXPERT_MODE_DESCRIPTION = _(
    'Expert mode enables advanced features and relaxes guardrails. '
    'Use extra caution in expert mode.')

# TRANSLATORS: Used both as (1) a checkbox label in the preferences dialog
# to bypass some safety guardails for advanced users and (2) the title of
# dialog asking to confirm this choice.
EXPERT_MODE_MSG = _('Expert mode')

# TRANSLATORS: Notice shown in an infobar after changing a preference
# that requires starting the application to be effective.
RESTART_APP_MSG = _("Restart BleachBit for full effect.")


class PreferencesDialog:

    """Present the preferences dialog and save changes"""

    def __init__(self, parent, cb_refresh_operations, cb_set_windows10_theme):
        self.cb_refresh_operations = cb_refresh_operations
        self.cb_set_windows10_theme = cb_set_windows10_theme

        self.parent = parent
        # TRANSLATORS: Title for the preferences dialog
        self.dialog = Gtk.Dialog(title=_("Preferences"),
                                 transient_for=parent,
                                 modal=True,
                                 destroy_with_parent=True)
        self.dialog.set_default_size(760, 520)

        self._locations_notice_css_provider = None
        self._default_options_box = None
        self._cookie_page_loaded = False
        self._cookie_page_container = None

        # Add InfoBar for non-blocking messages
        self.infobar = Gtk.InfoBar()
        self.infobar.set_show_close_button(True)
        self.infobar.connect('response', self._on_infobar_response)
        self.infobar_label = Gtk.Label()
        self.infobar_label.set_line_wrap(True)
        self.infobar.get_content_area().add(self.infobar_label)
        self.dialog.get_content_area().pack_start(self.infobar, False, False, 0)
        self._infobar_timeout_id = None

        content_box = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL, spacing=18)
        content_box.set_border_width(12)

        self.page_stack = Gtk.Stack()
        self.page_stack.set_transition_type(Gtk.StackTransitionType.NONE)
        self.page_stack.set_hhomogeneous(False)
        self.page_stack.set_vhomogeneous(False)
        self.page_stack.connect('notify::visible-child-name',
                                self.__on_stack_page_changed)

        sidebar = Gtk.StackSidebar()
        sidebar.set_stack(self.page_stack)
        sidebar.set_vexpand(True)
        sidebar.set_margin_end(6)

        pages = [
            # TRANSLATORS: Sidebar label for the general settings page of the preferences dialog.
            (self.__general_page, 'general', _("General"), True),
            # TRANSLATORS: Sidebar label for the software updates page of the preferences dialog.
            (self.__updates_page, 'updates', _("Updates"),
             online_update_notification_enabled),
            # TRANSLATORS: Sidebar label for the "Languages" page in the Preferences dialog.
            (self.__languages_page, 'languages', _("Languages"), True),
            # TRANSLATORS: Sidebar label for the page managing browser cookies in
            # the Preferences dialog.
            (self.__cookies_page, 'cookies', _("Cookies"), True),
            (lambda: self.__locations_page(LOCATIONS_WHITELIST),
             # TRANSLATORS: Sidebar label for the "keep list" (whitelist) page of
             # the preferences dialog. This page lists paths that will be preserved
             # (not deleted) during cleaning.
             'keep-list', _("Keep list"), True),
            (lambda: self.__locations_page(LOCATIONS_CUSTOM),
             # TRANSLATORS: Sidebar label for the page where users add their own paths to clean.
             # Short label meaning "Custom locations".
             'custom', _("Custom"), True),
            # TRANSLATORS: Sidebar label for the drives page of the preferences dialog.
            (self.__drives_page, 'drives', _("Drives"), True),
        ]

        for page_func, page_id, page_title, condition in pages:
            if condition:
                self.page_stack.add_titled(page_func(), page_id, page_title)

        content_box.pack_start(sidebar, False, False, 0)
        content_box.pack_start(self.page_stack, True, True, 0)

        # pack_start parameters: child, expand (reserve space), fill (actually fill it), padding
        self.dialog.get_content_area().pack_start(content_box, True, True, 0)
        self.dialog.add_button(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)

        self.refresh_operations = False

    def select_page(self, page_name):
        """Show the requested preferences page."""
        if not page_name:
            return
        self.page_stack.set_visible_child_name(page_name)
        if page_name == 'cookies':
            self.__ensure_cookie_page()

    def _create_checkbox(self, label, option_id, vbox=None, tooltip=None,
                         requires_option=None, store_as_attr=None):
        """Helper to create a checkbox with common patterns.

        Args:
            label: The checkbox label text
            option_id: The option ID to connect to
            vbox: Optional vbox to pack the checkbox into
            tooltip: Optional tooltip text
            requires_option: Optional option ID that must be enabled for sensitivity
            store_as_attr: Optional attribute name to store checkbox as self.{name}

        Returns:
            The created Gtk.CheckButton
        """
        cb = Gtk.CheckButton(label=label)
        cb.set_active(options.get(option_id))
        cb.connect('toggled', self.__toggle_callback, option_id)

        if tooltip:
            cb.set_tooltip_text(tooltip)

        if requires_option is not None and not options.get(requires_option):
            cb.set_sensitive(False)
        elif options.has_override(option_id):
            cb.set_sensitive(False)

        if store_as_attr:
            setattr(self, store_as_attr, cb)

        if vbox is None:
            vbox = self._default_options_box
        vbox.pack_start(cb, False, True, 0)

        return cb

    def __del__(self):
        """Destructor called when the dialog is closing"""
        if self.refresh_operations:
            # refresh the list of cleaners
            self.cb_refresh_operations()

    def _on_infobar_response(self, _infobar, _response_id):
        """Handle InfoBar close button click"""
        if self._infobar_timeout_id:
            GLib.source_remove(self._infobar_timeout_id)
            self._infobar_timeout_id = None
        self.infobar.hide()

    def _hide_infobar(self):
        """Hide the InfoBar (used for auto-dismiss timeout)"""
        self._infobar_timeout_id = None
        self.infobar.hide()
        return False  # Remove from GLib timeout

    def show_infobar(self, message, message_type=Gtk.MessageType.ERROR):
        """Show a non-blocking InfoBar message that auto-dismisses

        Args:
            message: The message to display
            message_type: Gtk.MessageType (ERROR, WARNING, INFO, etc.)
        """
        # Cancel any existing timeout
        if self._infobar_timeout_id:
            GLib.source_remove(self._infobar_timeout_id)
            self._infobar_timeout_id = None
        self.infobar_label.set_text(message)
        self.infobar.set_message_type(message_type)
        self.infobar.show_all()
        self._infobar_timeout_id = GLib.timeout_add_seconds(
            15, self._hide_infobar)

    def __on_expert_mode_toggled(self, cb):
        """Callback for expert mode checkbox"""
        new_value = cb.get_active()
        if new_value:
            from bleachbit.GuiBasic import warning_confirm_dialog
            confirmed, _remember_choice = warning_confirm_dialog(
                self.dialog,
                EXPERT_MODE_MSG,
                EXPERT_MODE_DESCRIPTION,
                show_checkbox=False
            )
            if not confirmed:
                cb.set_active(False)
                return
        options.set('expert_mode', new_value)
        self.reset_warnings_button.set_sensitive(new_value)
        if hasattr(self, 'cb_delete_confirmation'):
            self.cb_delete_confirmation.set_sensitive(new_value)
        if hasattr(self, 'cb_refresh_operations') and self.cb_refresh_operations:
            self.refresh_operations = True

    def __toggle_callback(self, _cell, path):
        """Callback function to toggle option"""
        options.toggle(path)
        if online_update_notification_enabled:
            self.cb_beta.set_sensitive(options.get('check_online_updates'))
            if 'nt' == os.name:
                self.cb_winapp2.set_sensitive(
                    options.get('check_online_updates'))
        if 'auto_hide' == path:
            self.refresh_operations = True
        if 'dark_mode' == path:
            logger.debug("Toggling dark mode to %s", options.get('dark_mode'))
            theme_widget = self.parent or self.dialog
            before_dark = detect_dark_background(theme_widget)
            if 'nt' == os.name and options.get('win10_theme'):
                self.cb_set_windows10_theme()

            settings = self.dialog.get_settings()
            if settings:
                settings.set_property(
                    'gtk-application-prefer-dark-theme', options.get('dark_mode'))
            else:
                logger.warning("Could not get GTK settings to apply dark mode")

            flush_gtk_events()
            after_dark = detect_dark_background(theme_widget)

            if not os.name == 'nt' and should_show_dark_mode_warning(before_dark, after_dark):
                self.show_infobar(
                    # TRANSLATORS: Notice shown in an infobar when toggling
                    # dark mode on Linux.
                    _("Some GTK themes do not support both light and dark modes."),
                    Gtk.MessageType.WARNING)
        if 'win10_theme' == path:
            self.cb_set_windows10_theme()
        if 'debug' == path:
            from bleachbit.Log import set_root_log_level
            set_root_log_level(options.get('debug'))
        if 'kde_shred_menu_option' == path:
            from bleachbit.DesktopMenuOptions import install_kde_service_menu_file
            install_kde_service_menu_file()
        if 'use_fontconfig_backend' == path:
            self.show_infobar(RESTART_APP_MSG, Gtk.MessageType.INFO)

    def __reset_warning_preferences(self, _button):
        """Reset saved warning confirmations."""
        options.clear_warning_preferences()
        self.show_infobar(
            # TRANSLATORS: Success message shown in the infobar.
            _("Warning confirmations reset."),
            Gtk.MessageType.INFO)

    def __create_update_widgets(self, vbox):
        """Create and configure update-related checkboxes."""
        if not online_update_notification_enabled:
            return
        self._create_checkbox(
            # TRANSLATORS: Checkbox label in the preferences dialog.
            _("Check periodically for software updates via the Internet"),
            'check_online_updates',
            vbox=vbox,
            # TRANSLATORS: Tooltip for the online updates checkbox in the preferences dialog.
            tooltip=_("If an update is found, you will be given the option to view information about it.  Then, you may manually download and install the update."))

        updates_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        updates_box.set_margin_start(18)

        self._create_checkbox(
            # TRANSLATORS: Checkbox label in the preferences dialog.
            _("Check for new beta releases"),
            'check_beta',
            vbox=updates_box,
            requires_option='check_online_updates',
            store_as_attr='cb_beta')

        if 'nt' == os.name:
            self._create_checkbox(
                # TRANSLATORS: Checkbox label in the preferences dialog.
                # Winapp2.ini is a set of cleaning rules from this project:
                # https://github.com/MoscaDotTo/Winapp2
                _("Download and update cleaners from community (winapp2.ini)"),
                'update_winapp2',
                vbox=updates_box,
                requires_option='check_online_updates',
                store_as_attr='cb_winapp2')
        vbox.pack_start(updates_box, False, True, 0)

    def __create_language_widgets(self, vbox):
        """Create and configure language selection widgets."""
        lang_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        is_auto_detect = options.get("auto_detect_lang")
        # TRANSLATORS: Checkbutton label in the preferences dialog.
        self.cb_auto_lang = Gtk.CheckButton(label=_("Auto-detect language"))
        self.cb_auto_lang.set_active(is_auto_detect)
        self.cb_auto_lang.set_tooltip_text(
            # TRANSLATORS: Tooltip explaining the auto-detect language option.
            _("Automatically detect the system language"))
        lang_box.pack_start(self.cb_auto_lang, False, True, 0)

        self.lang_select_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        # TRANSLATORS: Label for the language selection dropdown.
        lang_label = Gtk.Label(label=_("Language:"))
        lang_label.set_margin_start(20)  # Add some indentation
        self.lang_select_box.pack_start(lang_label, False, True, 5)

        self.lang_combo = Gtk.ComboBoxText()
        current_lang_code = get_active_language_code()
        # Add available languages
        lang_idx = 0
        active_language_idx = None
        try:
            supported_langs = get_supported_language_code_name_dict().items()
        except (KeyError, ValueError, Exception) as e:
            logger.error("Failed to get list of supported languages: %s", e)
            supported_langs = [('en_us', 'English')]
        for lang_code, native in supported_langs:
            if native:
                self.lang_combo.append_text(f"{native} ({lang_code})")
            else:
                self.lang_combo.append_text(lang_code)
            if lang_code == current_lang_code:
                active_language_idx = lang_idx
            lang_idx += 1
        if active_language_idx is not None:
            self.lang_combo.set_active(active_language_idx)
        # set_wrap_width() prevents infinite space to scroll up.
        # https://github.com/bleachbit/bleachbit/issues/1764
        self.lang_combo.set_wrap_width(1)
        self.lang_select_box.pack_start(self.lang_combo, False, True, 0)
        lang_box.pack_start(self.lang_select_box, False, True, 0)

        vbox.pack_start(lang_box, False, True, 0)
        self.lang_select_box.set_sensitive(not is_auto_detect)
        self.cb_auto_lang.connect('toggled', self.on_auto_detect_toggled)
        self.lang_combo.connect('changed', self.on_lang_changed)

    def on_lang_changed(self, widget):
        """Callback for when the language combobox is changed."""
        text = widget.get_active_text()
        # Extract language code from the format "Native Name (lang_code)"
        lang_code = text.split("(")[-1].rstrip(")")
        if lang_code:
            options.set("forced_language", lang_code, section="bleachbit")
        else:
            logger.warning(
                "No language code found in combobox for text %s", text)
        setup_translation()
        self.refresh_operations = True
        self.show_infobar(RESTART_APP_MSG, Gtk.MessageType.INFO)

    def on_auto_detect_toggled(self, _widget):
        """Callback for when the auto-detect language checkbox is toggled."""
        self.__toggle_callback(None, 'auto_detect_lang')
        is_auto_detect = options.get("auto_detect_lang")
        self.lang_select_box.set_sensitive(not is_auto_detect)
        if is_auto_detect:
            options.set("forced_language", "", section="bleachbit")
        setup_translation()
        self.refresh_operations = True
        self.show_infobar(RESTART_APP_MSG, Gtk.MessageType.INFO)

    def __create_general_checkboxes(self, vbox):
        """Create and configure general checkboxes."""
        # TRANSLATORS: Overwriting is the same as shredding.  It is a way
        # to prevent recovery of the data. You could also translate
        # 'Shred files to prevent recovery.'
        self._create_checkbox(
            # TRANSLATORS: Checkbox label in the preferences dialog.
            _("Overwrite contents of files to prevent recovery"),
            'shred',
            vbox=vbox,
            # TRANSLATORS: Tooltip for the shred checkbox in the preferences dialog.
            tooltip=_("Overwriting is ineffective on some file systems and with certain BleachBit operations.  Overwriting is significantly slower."))

        self._create_checkbox(
            # TRANSLATORS: Checkbox label in the preferences dialog.
            _("Exit after cleaning"),
            'exit_done',
            vbox=vbox)

        self._create_checkbox(
            # TRANSLATORS: Checkbox label in the preferences dialog.
            # Ask for confirmation before deleting items.
            _("Confirm before delete"),
            'delete_confirmation',
            vbox=vbox,
            tooltip=REQUIRES_EXPERT_MODE,
            requires_option='expert_mode',
            store_as_attr='cb_delete_confirmation')

        self.reset_warnings_button = Gtk.Button.new_with_label(
            # TRANSLATORS: Button label in the preferences. 'Reset' is a verb.
            label=_("Reset warning confirmations"))
        self.reset_warnings_button.set_halign(Gtk.Align.START)
        self.reset_warnings_button.set_margin_top(6)
        self.reset_warnings_button.set_tooltip_text(REQUIRES_EXPERT_MODE)
        self.reset_warnings_button.connect(
            'clicked', self.__reset_warning_preferences)
        self.reset_warnings_button.set_sensitive(options.get('expert_mode'))
        vbox.pack_start(self.reset_warnings_button, False, True, 0)

        # TRANSLATORS: This means to hide cleaners which would do
        # nothing.  For example, if Firefox were never used on
        # this system, this option would hide Firefox to simplify
        # the list of cleaners.
        self._create_checkbox(
            _("Hide irrelevant cleaners"),
            'auto_hide',
            vbox=vbox)

        self._create_checkbox(
            # TRANSLATORS: Checkbox label in the preferences.
            _("Use IEC sizes (1 KiB = 1024 bytes) instead of SI (1 kB = 1000 bytes)"),
            "units_iec",
            vbox=vbox)

    def __create_page_box(self):
        """Create a standard page container box with consistent spacing and padding."""
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=18)
        vbox.set_border_width(12)
        return vbox

    def __create_section(self, parent, title):
        """Create a titled section with a bold heading and indented body."""
        section = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
        heading = Gtk.Label()
        heading.set_markup('<b>%s</b>' % GLib.markup_escape_text(title))
        heading.set_xalign(0)
        section.pack_start(heading, False, False, 0)

        body = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
        body.set_margin_start(12)
        section.pack_start(body, False, False, 0)
        parent.pack_start(section, False, False, 0)
        return body

    def __updates_page(self):
        """Return widget containing the updates page."""
        page = self.__create_page_box()

        # TRANSLATORS: Section title on the preferences updates page.
        updates_box = self.__create_section(page, _("Update notifications"))
        self.__create_update_widgets(updates_box)
        return page

    def __cookies_page(self):
        """Return widget containing the cookies page (lazy-loaded)."""
        page = self.__create_page_box()

        self._cookie_page_container = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL, spacing=0)
        page.pack_start(self._cookie_page_container, True, True, 0)
        return page

    def __ensure_cookie_page(self):
        """Lazy-load the cookie manager pane when first accessed."""
        if self._cookie_page_loaded:
            return
        cookie_page = CookieManagerPane()
        cookie_page.set_border_width(12)
        self._cookie_page_container.pack_start(cookie_page, True, True, 0)
        self._cookie_page_container.show_all()
        self._cookie_page_loaded = True

    def __on_stack_page_changed(self, stack, _param):
        """Callback when the visible page in the stack changes."""
        if stack.get_visible_child_name() == 'cookies':
            self.__ensure_cookie_page()

    def __general_page(self):
        """Return a widget containing the general page"""

        page = self.__create_page_box()

        # TRANSLATORS: Section title (noun) on the Preferences - General page
        # for options related to cleaning stored data.
        cleaning_box = self.__create_section(page, _("Cleaning"))
        self._default_options_box = cleaning_box
        self.__create_general_checkboxes(cleaning_box)

        # TRANSLATORS: Section title in Preferences for user interface options
        # such as dark mode and window behavior.
        interface_box = self.__create_section(page, _("Interface"))
        self._create_checkbox(
            # TRANSLATORS: Checkbox label to remember the window size and position
            # between application launches.
            _("Remember window geometry"),
            'remember_geometry',
            vbox=interface_box)

        self._create_checkbox(
            # TRANSLATORS: Checkbox label to enable dark mode in the UI.
            _("Dark mode"),
            'dark_mode',
            vbox=interface_box)

        if 'nt' == os.name:
            self._create_checkbox(
                # TRANSLATORS: Checkbox label to use the Windows 10-style visual theme
                # for the application interface.
                _("Windows 10 theme"),
                'win10_theme',
                vbox=interface_box)

            self._create_checkbox(
                # TRANSLATORS: Checkbox label to use the Fontconfig text rendering backend
                # instead of the default Windows renderer. The option may improve blurry text.
                # Fontconfig, which is software, is a proper noun.
                _("Use Fontconfig text rendering backend"),
                'use_fontconfig_backend',
                vbox=interface_box,
                # TRANSLATORS: Tooltip for the fontconfig text rendering checkbox in the preferences dialog.
                tooltip=_("May fix blurry or unreadable text. Requires restart."))

        integration_box = self.__create_section(
            # TRANSLATORS: Section title on the preferences general page,
            # labelling options for OS integration and advanced/developer features.
            page, _("Integration and advanced"))

        if 'nt' != os.name:
            self._create_checkbox(
                # TRANSLATORS: Checkbox label in the preferences dialog.
                # 'Shred' means securely delete a file to prevent data recovery.
                # 'Context menu' is the menu shown on right-click.
                # 'KDE Plasma' is a Linux desktop environment (proper noun, do not translate).
                _("Add the shred context menu to KDE Plasma"),
                'kde_shred_menu_option',
                vbox=integration_box)

        self._create_checkbox(
            # TRANSLATORS: Checkbox label to show diagnostic log messages for
            # troubleshooting.
            _("Show debug messages"),
            'debug',
            vbox=integration_box)

        self._create_checkbox(
            EXPERT_MODE_MSG,
            'expert_mode',
            vbox=integration_box,
            tooltip=EXPERT_MODE_DESCRIPTION,
            store_as_attr='cb_expert')
        self.cb_expert.connect('toggled', self.__on_expert_mode_toggled)

        return page

    def __drives_page(self):
        """Return widget containing the drives page"""

        def add_drive_cb(button):
            """Callback for adding a drive"""
            # TRANSLATORS: Title of a folder chooser dialog.
            title = _("Choose a folder")
            pathname = GuiBasic.browse_folder(
                self.parent, title, multiple=False, stock_button=Gtk.STOCK_ADD)
            if pathname:
                liststore.append([pathname])
                pathnames.append(pathname)
                options.set_list('shred_drives', pathnames)

        def remove_drive_cb(button):
            """Callback for removing a drive"""
            treeselection = treeview.get_selection()
            (model, _iter) = treeselection.get_selected()
            if None == _iter:
                # nothing selected
                return
            pathname = model[_iter][0]
            liststore.remove(_iter)
            pathnames.remove(pathname)
            options.set_list('shred_drives', pathnames)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=12)
        vbox.set_border_width(12)

        # TRANSLATORS: 'empty' means 'unallocated'.
        # The user should decide which drives to wipe. Then, for each
        # chosen drive, the user should specify one folder per drive.
        # For example, to wipe C:\, the user may choose %TEMP% which
        # is a folder anywhere on C:.
        drive_instruction_label = _("Choose a writable folder for each drive for "
                                    "which to wipe empty space.")
        notice = Gtk.Label(label=drive_instruction_label)
        notice.set_line_wrap(True)
        notice.set_xalign(0.0)
        notice.set_margin_start(12)
        notice.set_margin_end(12)
        vbox.pack_start(notice, False, True, 0)

        notice2 = Gtk.Label(label=EMPTY_SPACE_WARNING)
        notice2.set_line_wrap(True)
        notice2.set_xalign(0.0)
        notice2.set_margin_start(12)
        notice2.set_margin_end(12)
        vbox.pack_start(notice2, False, True, 0)

        liststore = Gtk.ListStore(str)

        pathnames = options.get_list('shred_drives')
        if pathnames:
            pathnames = sorted(pathnames)
        if not pathnames:
            pathnames = []
        for pathname in pathnames:
            liststore.append([pathname])
        treeview = Gtk.TreeView.new_with_model(liststore)
        crt = Gtk.CellRendererText()
        tvc = Gtk.TreeViewColumn(None, crt, text=0)
        treeview.append_column(tvc)

        vbox.pack_start(treeview, True, True, 0)

        # TRANSLATORS: In the preferences dialog, this button adds a path to
        # the list of paths
        button_add = Gtk.Button.new_with_label(label=_p('button', 'Add'))
        button_add.connect("clicked", add_drive_cb)
        # TRANSLATORS: In the preferences dialog, this button removes a path
        # from the list of paths
        button_remove = Gtk.Button.new_with_label(label=_p('button', 'Remove'))
        button_remove.connect("clicked", remove_drive_cb)

        button_box = Gtk.ButtonBox(orientation=Gtk.Orientation.HORIZONTAL)
        button_box.set_layout(Gtk.ButtonBoxStyle.START)
        button_box.pack_start(button_add, True, True, 0)
        button_box.pack_start(button_remove, True, True, 0)
        vbox.pack_start(button_box, False, True, 0)

        return vbox

    def __languages_page(self):
        """Return widget containing the languages page"""

        def keep_lang_toggled_cb(cell, path, liststore):
            """Callback for toggling the 'keep' column"""
            __iter = liststore.get_iter_from_string(path)
            value = not liststore.get_value(__iter, 0)
            liststore.set(__iter, 0, value)
            langid = liststore[path][1]
            options.set_language(langid, value)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=12)
        vbox.set_border_width(12)

        ui_language_box = self.__create_section(
            vbox,
            # TRANSLATORS: Section title on the preferences languages page.
            _("BleachBit interface language"))

        self.__create_language_widgets(ui_language_box)

        # Windows does not have locale cleaner.
        if 'posix' != os.name:
            return vbox

        cleanup_box = self.__create_section(
            vbox,
            # TRANSLATORS: Section title on the preferences languages page.
            _("Language files to keep when cleaning applications"))

        # populate data
        liststore = Gtk.ListStore('gboolean', str, str)
        for lang, native in get_supported_language_code_name_dict().items():
            liststore.append([(options.get_language(lang)), lang, native])

        # create treeview
        treeview = Gtk.TreeView.new_with_model(liststore)

        # create column views
        self.renderer0 = Gtk.CellRendererToggle()
        self.renderer0.set_property('activatable', True)
        self.renderer0.connect('toggled', keep_lang_toggled_cb, liststore)
        # TRANSLATORS: Column header in the languages treeview.
        # This column controls whether to keep the language.
        self.column0 = Gtk.TreeViewColumn(_("Keep"),
                                          self.renderer0, active=0)
        treeview.append_column(self.column0)

        self.renderer1 = Gtk.CellRendererText()
        # TRANSLATORS: Column header in the languages treeview showing the
        # language code (e.g., 'en_US').
        self.column1 = Gtk.TreeViewColumn(_("Code"), self.renderer1, text=1)
        treeview.append_column(self.column1)

        self.renderer2 = Gtk.CellRendererText()
        # TRANSLATORS: Column header in the languages treeview showing the
        # native name of the language.
        self.column2 = Gtk.TreeViewColumn(_("Name"), self.renderer2, text=2)
        treeview.append_column(self.column2)
        treeview.set_search_column(2)

        # finish
        swindow = Gtk.ScrolledWindow()
        swindow.set_overlay_scrolling(False)
        swindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        swindow.set_size_request(300, 200)
        swindow.add(treeview)
        cleanup_box.pack_start(swindow, True, True, 0)
        return vbox

    def _check_path_exists(self, pathname):
        """Check if a path exists in either keep lists or custom lists
        Returns True if path exists, False otherwise"""
        whitelist_paths = options.get_whitelist_paths()
        custom_paths = options.get_custom_paths()

        # Check in whitelist
        for path in whitelist_paths:
            if pathname == path[1]:
                # TRANSLATORS: Error message shown in the infobar.
                msg = _("This path already exists in the keep list.")
                self.show_infobar(msg, Gtk.MessageType.ERROR)
                return True

        # Check in custom
        for path in custom_paths:
            if pathname == path[1]:
                # TRANSLATORS: Error message shown in the infobar.
                msg = _("This path already exists in the custom list.")
                self.show_infobar(msg, Gtk.MessageType.ERROR)
                return True

        return False

    def _check_protected_path(self, pathname):
        """Check if path is protected and warn user if so.

        Returns True if it's safe to proceed, False if user cancelled.
        """
        logger.debug("Checking protected path: %s", pathname)
        match_info = ProtectedPath.check_protected_path(pathname)
        if match_info is None:
            return True

        if not options.get('expert_mode'):
            self.show_infobar(
                # TRANSLATORS: Error message shown in the infobar.
                _("This path is protected. To bypass protection, enable expert mode."),
                Gtk.MessageType.WARNING)
            return False

        # Check if user already confirmed this path
        normalized = ProtectedPath._normalize_for_comparison(
            pathname, match_info['case_sensitive'])
        warning_key = 'protected_path:' + normalized
        if options.get_warning_preference(warning_key):
            return True

        # Calculate impact
        # FIXME later: this can be very slow with many objects
        logger.debug("Checking protected path impact: %s", pathname)
        impact = ProtectedPath.calculate_impact(pathname)

        # Generate warning message
        warning_msg = ProtectedPath.get_warning_message(pathname, impact)

        # Show warning dialog
        confirmed, remember = GuiBasic.warning_confirm_dialog(
            self.dialog,
            # TRANSLATORS: Title of warning dialog shown when adding a protected path.
            _("Protected Path"),
            warning_msg
        )

        if confirmed and remember:
            options.remember_warning_preference(warning_key)

        return confirmed

    def _add_path(self, pathname, path_type, page_type, liststore, pathnames):
        """Common function to add a path to either whitelist or custom list"""
        if self._check_path_exists(pathname):
            return

        # Check for protected paths when adding to custom (delete) list
        if page_type == LOCATIONS_CUSTOM:
            if not self._check_protected_path(pathname):
                return

        # TRANSLATORS: Noun used as a column header in the preferences dialog.
        type_str_file = _('File')
        # TRANSLATORS: Noun used as a column header in the preferences dialog.
        type_str_folder = _('Folder')
        type_str = type_str_file if path_type == 'file' else type_str_folder
        display_path = pathname.encode('utf-8', errors='replace').decode('utf-8')
        liststore.append([type_str, display_path])
        pathnames.append([path_type, pathname])

        if page_type == LOCATIONS_WHITELIST:
            options.set_whitelist_paths(pathnames)
        else:
            options.set_custom_paths(pathnames)

        # TRANSLATORS: %s is a file or folder path that was just added
        self.show_infobar(_("Added: %s") % display_path,
                          Gtk.MessageType.INFO)

    def _remove_path(self, treeview, liststore, pathnames, page_type):
        """Common function to remove a path from either whitelist or custom list"""
        treeselection = treeview.get_selection()
        (model, _iter) = treeselection.get_selected()
        if None == _iter:
            return
        tree_path = model.get_path(_iter)
        row_index = tree_path.get_indices()[0]
        display_path = model[_iter][1]
        liststore.remove(_iter)
        pathnames.pop(row_index)
        if page_type == LOCATIONS_WHITELIST:
            options.set_whitelist_paths(pathnames)
        else:
            options.set_custom_paths(pathnames)
        # TRANSLATORS: %s is a file or folder path that was just removed
        self.show_infobar(_("Removed: %s") % display_path,
                          Gtk.MessageType.INFO)

    def __locations_page(self, page_type):
        """Return a widget containing a list of files and folders"""

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=12)
        vbox.set_border_width(12)

        # load data
        pathnames = []
        if LOCATIONS_WHITELIST == page_type:
            pathnames = options.get_whitelist_paths()
        elif LOCATIONS_CUSTOM == page_type:
            pathnames = options.get_custom_paths()
        else:
            raise RuntimeError("Invalid page type: '%s'" % page_type)
        liststore = Gtk.ListStore(str, str)
        for paths in pathnames:
            type_code = paths[0]
            type_str = None
            if type_code == 'file':
                type_str = _('File')
            elif type_code == 'folder':
                type_str = _('Folder')
            else:
                raise RuntimeError("Invalid type code: '%s'" % type_code)
            path = paths[1]
            display_path = path.encode('utf-8', errors='replace').decode('utf-8')
            liststore.append([type_str, display_path])

        if not self._locations_notice_css_provider:
            self._locations_notice_css_provider = Gtk.CssProvider()
            self._locations_notice_css_provider.load_from_data(b"""
                .bb-locations-notice-whitelist {
                    background-color: rgba(46, 139, 87, 0.12);
                    border-radius: 6px;
                    padding: 6px;
                }
                .bb-locations-notice-custom {
                    background-color: rgba(210, 120, 0, 0.12);
                    border-radius: 6px;
                    padding: 6px;
                }
            """)

        if LOCATIONS_WHITELIST == page_type:
            # TRANSLATORS: Notice label at the top of the keeplist (whitelist)
            # page in the preferences dialog.
            # "Paths" is used generically to refer to both files and folders.
            notice_text = _("These paths will not be deleted or modified.")
            notice_icon = "emblem-readonly"
            notice_class = "bb-locations-notice-whitelist"
        elif LOCATIONS_CUSTOM == page_type:
            # TRANSLATORS: Notice label at the top of the custom (delete)
            # page in the preferences dialog.
            notice_text = _("These locations can be selected for deletion.")
            notice_icon = "edit-delete"
            notice_class = "bb-locations-notice-custom"
        else:
            raise RuntimeError(f"Invalid page type: {page_type}")

        notice_label = Gtk.Label(label=notice_text)
        notice_label.set_line_wrap(True)
        notice_label.set_xalign(0.0)

        notice_image = Gtk.Image.new_from_icon_name(
            notice_icon, Gtk.IconSize.MENU)
        notice_image.set_valign(Gtk.Align.START)

        notice_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        notice_box.pack_start(notice_image, False, False, 0)
        notice_box.pack_start(notice_label, True, True, 0)

        notice_frame = Gtk.EventBox()
        notice_frame.set_visible_window(True)
        notice_frame.add(notice_box)
        notice_frame.set_margin_bottom(6)
        notice_frame.set_margin_top(6)
        style_context = notice_frame.get_style_context()
        style_context.add_provider(
            self._locations_notice_css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
        style_context.add_class(notice_class)
        vbox.pack_start(notice_frame, False, False, 0)

        # create treeview
        treeview = Gtk.TreeView.new_with_model(liststore)

        # create column views
        self.renderer0 = Gtk.CellRendererText()
        # TRANSLATORS: Column header in the preferences dialog showing whether
        # the item is a file or folder.
        self.column0 = Gtk.TreeViewColumn(_("Type"), self.renderer0, text=0)
        treeview.append_column(self.column0)

        self.renderer1 = Gtk.CellRendererText()
        # TRANSLATORS: In the tree view "Path" is used generically to refer to a
        # file, a folder, or a pattern describing either
        self.column1 = Gtk.TreeViewColumn(_("Path"), self.renderer1, text=1)
        treeview.append_column(self.column1)
        treeview.set_search_column(1)

        # finish tree view
        swindow = Gtk.ScrolledWindow()
        swindow.set_overlay_scrolling(False)
        swindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        swindow.set_size_request(300, 200)
        swindow.add(treeview)

        vbox.pack_start(swindow, True, True, 0)

        # buttons that modify the list
        def add_file_cb(button):
            """Callback for adding a file"""
            # TRANSLATORS: Title of a file chooser dialog.
            title = _("Choose a file")
            pathname = GuiBasic.browse_file(self.parent, title)
            if pathname:
                self._add_path(pathname, 'file', page_type,
                               liststore, pathnames)

        def add_folder_cb(button):
            """Callback for adding a folder"""
            # TRANSLATORS: Title of a folder chooser dialog.
            title = _("Choose a folder")
            pathname = GuiBasic.browse_folder(self.parent, title,
                                              multiple=False, stock_button=Gtk.STOCK_ADD)
            if pathname:
                self._add_path(pathname, 'folder', page_type,
                               liststore, pathnames)

        def remove_path_cb(button):
            """Callback for removing a path"""
            self._remove_path(treeview, liststore, pathnames, page_type)

        button_add_file = Gtk.Button.new_with_label(
            label=_p('button', 'Add file'))
        button_add_file.connect("clicked", add_file_cb)

        button_add_folder = Gtk.Button.new_with_label(
            label=_p('button', 'Add folder'))
        button_add_folder.connect("clicked", add_folder_cb)

        button_remove = Gtk.Button.new_with_label(label=_p('button', 'Remove'))
        button_remove.connect("clicked", remove_path_cb)

        button_box = Gtk.ButtonBox(orientation=Gtk.Orientation.HORIZONTAL)
        button_box.set_layout(Gtk.ButtonBoxStyle.START)
        button_box.pack_start(button_add_file, True, True, 0)
        button_box.pack_start(button_add_folder, True, True, 0)
        button_box.pack_start(button_remove, True, True, 0)
        vbox.pack_start(button_box, False, True, 0)

        # return page
        return vbox

    def run(self, page_name=None):
        """Run the dialog, optionally opening to a specific page.

        Args:
            page_name: Optional page ID to select (e.g., 'cookies', 'custom').
        """
        self.dialog.show_all()
        self.infobar.hide()
        self.select_page(page_name)
        self.dialog.run()
        self.dialog.destroy()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

r"""
Functionality specific to Microsoft Windows

The Windows Registry terminology can be confusing. Take for example
the reference
* HKCU\\Software\\BleachBit
* CurrentVersion

These are the terms:
* 'HKCU' is an abbreviation for the hive HKEY_CURRENT_USER.
* 'HKCU\Software\BleachBit' is the key name.
* 'Software' is a sub-key of HCKU.
* 'BleachBit' is a sub-key of 'Software.'
* 'CurrentVersion' is the value name.
* '0.5.1' is the value data.


"""

# standard imports
import atexit
import base64
import ctypes
import decimal
import errno
import glob
import hashlib
import logging
import ntpath
import os
import pathlib
import re
import shutil
import struct
import sys
import threading
import time
import uuid
import xml.dom.minidom
from ctypes import wintypes
from decimal import Decimal
from pathlib import Path
from threading import Thread, Event
from uuid import UUID

# first party imports
import bleachbit
from bleachbit import FileUtilities, General, ARCH_BITS, IS_WINDOWS
from bleachbit.Language import get_text as _

if 'win32' == sys.platform:
    import winreg
    import pywintypes
    import win32api
    import win32clipboard
    import win32con
    import win32file
    import win32gui
    import win32process
    import win32security
    import win32service
    import win32serviceutil

    # Ensure GetClassInfo exists for compatibility and testing
    # Some win32gui builds don't have GetClassInfo, so we create a stub
    # In the future, consider GetClassInfoEx instead.
    def _get_class_info_fallback(hInstance, className):
        """Fallback GetClassInfo - returns a default atom value"""
        return (1234,)  # Return tuple with default atom

    # Force GetClassInfo to exist on the module for mocking compatibility
    # This ensures tests can patch it even if it doesn't exist natively
    setattr(win32gui, 'GetClassInfo', getattr(
        win32gui, 'GetClassInfo', _get_class_info_fallback))

    from ctypes import windll, byref
    from win32com.shell import shell, shellcon

    psapi = windll.psapi
    kernel = windll.kernel32

logger = logging.getLogger(__name__)

IO_REPARSE_TAG_MOUNT_POINT = 0xA0000003
FILE_ATTRIBUTE_REPARSE_POINT = 0x400

SPLASH_ICON_SIZE_PX = 256  # 256x256 pixels
SPLASH_CLOSE_TIMEOUT_MS = 1000

WINDOWS_SYSTEM_VAR = 'WindowsSystem'

_delete_parent_lock_admin = None
_delete_parent_lock_handle = None
_delete_parent_lock_key = None


class _POINT(ctypes.Structure):
    _fields_ = [('x', wintypes.LONG), ('y', wintypes.LONG)]


class _SIZE(ctypes.Structure):
    _fields_ = [('cx', wintypes.LONG), ('cy', wintypes.LONG)]


class _BLENDFUNCTION(ctypes.Structure):
    _fields_ = [
        ('BlendOp', wintypes.BYTE),
        ('BlendFlags', wintypes.BYTE),
        ('SourceConstantAlpha', wintypes.BYTE),
        ('AlphaFormat', wintypes.BYTE),
    ]


class _BITMAPINFOHEADER(ctypes.Structure):
    _fields_ = [
        ('biSize', wintypes.DWORD),
        ('biWidth', wintypes.LONG),
        ('biHeight', wintypes.LONG),
        ('biPlanes', wintypes.WORD),
        ('biBitCount', wintypes.WORD),
        ('biCompression', wintypes.DWORD),
        ('biSizeImage', wintypes.DWORD),
        ('biXPelsPerMeter', wintypes.LONG),
        ('biYPelsPerMeter', wintypes.LONG),
        ('biClrUsed', wintypes.DWORD),
        ('biClrImportant', wintypes.DWORD),
    ]


class _BITMAPINFO(ctypes.Structure):
    _fields_ = [('bmiHeader', _BITMAPINFOHEADER),
                ('bmiColors', wintypes.DWORD * 3)]


_BI_RGB = 0
_DIB_RGB_COLORS = 0
_AC_SRC_OVER = 0
_AC_SRC_ALPHA = 1
_ULW_ALPHA = 2


def get_splash_screen_delay_seconds():
    value = os.environ.get('BLEACHBIT_SPLASH_SCREEN_DELAY')
    if not value:
        return 0.0
    try:
        delay = float(value)
    except ValueError:
        logger.warning(
            'invalid BLEACHBIT_SPLASH_SCREEN_DELAY value: %r', value)
        return 0.0
    if delay < 0:
        logger.warning(
            'negative BLEACHBIT_SPLASH_SCREEN_DELAY value: %r', value)
        return 0.0
    return delay


def _get_environ_case_insensitive(environ, varname):
    """Get a Windows environment variable from a mapping."""
    if varname in environ:
        return environ[varname]
    varname_lower = varname.lower()
    for key, value in environ.items():
        if key.lower() == varname_lower:
            return value
    return None


def _is_64_bit_windows(environ):
    """Return whether the environment indicates a 64-bit Windows OS."""
    if _get_environ_case_insensitive(environ, 'ProgramW6432'):
        return True
    if _get_environ_case_insensitive(environ, 'PROCESSOR_ARCHITEW6432'):
        return True
    arch = _get_environ_case_insensitive(environ, 'PROCESSOR_ARCHITECTURE')
    return bool(arch and arch.upper() in ('AMD64', 'ARM64', 'IA64'))


def get_windows_system_paths(environ=None, process_bits=None):
    """Return Windows system directories for native and 32-bit targets.

    On a 32-bit process running on 64-bit Windows, System32 is redirected to
    SysWOW64, so Sysnative is used for the native 64-bit system directory.

    Returns:
        list: A list of Windows system directory paths. On 64-bit Windows,
              this typically includes both the native system directory
              (System32 or Sysnative) and SysWOW64. On 32-bit Windows,
              only System32 is returned. Returns an empty list if the
              Windows directory cannot be determined.

    """
    environ = environ or os.environ
    process_bits = process_bits or ARCH_BITS
    windir = (_get_environ_case_insensitive(environ, 'WinDir') or
              _get_environ_case_insensitive(environ, 'SystemRoot'))
    if not windir:
        return []
    windir = windir.rstrip('\\/')
    if _is_64_bit_windows(environ):
        native = 'Sysnative' if process_bits == 32 else 'System32'
        paths = (ntpath.join(windir, native), ntpath.join(windir, 'SysWOW64'))
    else:
        paths = (ntpath.join(windir, 'System32'), )

    return list(paths)


def expand_windows_system_vars(pathname, system_paths=None):
    """Expand the multi-value %WindowsSystem% variable in a path.

    Args:
        pathname: The path string containing the %WindowsSystem% variable.
        system_paths: Optional list of system paths to use. If not provided,
                     the function will determine the appropriate system paths
                     based on the current environment.

    Returns:
        list: A list of expanded path strings, one for each system path.
    """
    pattern = re.compile(rf'%{WINDOWS_SYSTEM_VAR}%', flags=re.IGNORECASE)
    if not pattern.search(pathname):
        return [pathname]
    system_paths = system_paths or get_windows_system_paths()
    if not system_paths:
        return [pathname]
    return [
        pattern.sub(lambda _match: system_path, pathname)
        for system_path in system_paths
    ]


def browse_file(_, title):
    """Ask the user to select a single file.  Return full path"""
    try:
        ret = win32gui.GetOpenFileNameW(None,
                                        Flags=win32con.OFN_EXPLORER
                                        | win32con.OFN_FILEMUSTEXIST
                                        | win32con.OFN_HIDEREADONLY,
                                        Title=title)
    except pywintypes.error as e:
        logger = logging.getLogger(__name__)
        if 0 == e.winerror:
            logger.debug('browse_file(): user cancelled')
        else:
            logger.exception('exception in browse_file()')
        return None
    return ret[0]


def browse_files(_, title):
    """Ask the user to select files.  Return full paths"""
    try:
        # The File parameter is a hack to increase the buffer length.
        ret = win32gui.GetOpenFileNameW(None,
                                        File='\x00' * 10240,
                                        Flags=win32con.OFN_ALLOWMULTISELECT
                                        | win32con.OFN_EXPLORER
                                        | win32con.OFN_FILEMUSTEXIST
                                        | win32con.OFN_HIDEREADONLY,
                                        Title=title)
    except pywintypes.error as e:
        if 0 == e.winerror:
            logger.debug('browse_files(): user cancelled')
        else:
            logger.exception('exception in browse_files()')
        return None
    _split = ret[0].split('\x00')
    if 1 == len(_split):
        # only one filename
        return _split
    dirname = _split[0]
    pathnames = [os.path.join(dirname, fname) for fname in _split[1:]]
    return pathnames


def browse_folder(_, title):
    """Ask the user to select a folder.  Return full path."""
    flags = 0x0010  # SHBrowseForFolder path input
    pidl = shell.SHBrowseForFolder(None, None, title, flags)[0]
    if pidl is None:
        # user cancelled
        return None
    fullpath = shell.SHGetPathFromIDListW(pidl)
    return fullpath


def cleanup_nonce():
    """On exit, clean up GTK junk files"""
    for fn in glob.glob(os.path.expandvars(r'%TEMP%\gdbus-nonce-file-*')):
        logger.debug('cleaning GTK nonce file: %s', fn)
        FileUtilities.delete(fn)


def csidl_to_environ(varname, csidl):
    """Define an environment variable from a CSIDL for use in CleanerML and Winapp2.ini"""
    try:
        sppath = shell.SHGetSpecialFolderPath(None, csidl)
    except:
        logger.info(
            'exception when getting special folder path for %s', varname)
        return
    # there is exception handling in set_environ()
    set_environ(varname, sppath)


def _delete_parent_lock_needed(pathname):
    """
    Check if a parent directory lock is needed.

    This is only needed on Windows for administrator users
    when the path is not in the user's profile directory.
    """
    if os.name != 'nt':
        return False
    global _delete_parent_lock_admin
    if _delete_parent_lock_admin is None:
        try:
            _delete_parent_lock_admin = shell.IsUserAnAdmin()
        except Exception:
            logger.exception(
                'error checking whether current user is an administrator')
            _delete_parent_lock_admin = True
    return _delete_parent_lock_admin and not _path_in_user_profile(pathname)


def _path_for_comparison(pathname):
    """
    Normalize a path for comparison.
    """
    return os.path.normcase(os.path.abspath(
        FileUtilities.extended_path_undo(pathname)))


def _path_in_user_profile(pathname):
    """
    Check if a path is within the user's profile directory.
    """
    userprofile = os.environ.get('USERPROFILE')
    if not userprofile:
        return False
    try:
        profile_path = _path_for_comparison(userprofile)
        compare_path = _path_for_comparison(pathname)
        return os.path.commonpath((profile_path, compare_path)) == profile_path
    except (OSError, ValueError):
        return False


def _delete_parent_directory(pathname):
    """
    Get the parent directory of a file or directory.
    """
    path = os.path.abspath(FileUtilities.extended_path_undo(pathname))
    return FileUtilities.extended_path(os.path.dirname(path))


def _close_delete_parent_lock():
    """Close the parent lock handle."""
    global _delete_parent_lock_handle
    global _delete_parent_lock_key
    if _delete_parent_lock_handle is not None:
        logger.debug('Closing parent lock handle for %s',
                     _delete_parent_lock_key)
        win32file.CloseHandle(_delete_parent_lock_handle)
        _delete_parent_lock_handle = None
        _delete_parent_lock_key = None


def _lock_delete_parent(pathname):
    """
    Lock the parent directory of pathname to prevent it from being deleted.

    This function does not perform the deletion.
    """
    global _delete_parent_lock_handle
    global _delete_parent_lock_key
    parent = _delete_parent_directory(pathname)
    parent_key = os.path.normcase(parent)
    if _delete_parent_lock_handle is not None and _delete_parent_lock_key == parent_key:
        logger.debug('Reusing parent lock handle for %s', parent_key)
        return
    _close_delete_parent_lock()
    flags = win32con.FILE_FLAG_BACKUP_SEMANTICS | getattr(
        win32con, 'FILE_FLAG_OPEN_REPARSE_POINT', 0x00200000)
    access = getattr(win32con, 'FILE_READ_ATTRIBUTES', 0x80)
    try:
        handle = win32file.CreateFile(
            parent,
            access,
            win32con.FILE_SHARE_READ,
            None,
            win32con.OPEN_EXISTING,
            flags,
            None)
    except pywintypes.error as e:
        raise OSError(errno.EACCES,
                      "Access denied locking directory before delete()",
                      pathname) from e
    if handle == win32file.INVALID_HANDLE_VALUE:
        raise OSError(errno.EACCES,
                      "Access denied locking directory before delete()",
                      pathname)
    try:
        attrs = win32file.GetFileAttributesW(parent)
    except pywintypes.error:
        win32file.CloseHandle(handle)
        raise
    if attrs & FILE_ATTRIBUTE_REPARSE_POINT:
        win32file.CloseHandle(handle)
        raise OSError(errno.EACCES,
                      "Refusing to delete through a directory link",
                      pathname)
    logger.debug('Opened parent lock handle for %s', parent_key)
    _delete_parent_lock_handle = handle
    _delete_parent_lock_key = parent_key


def is_handle_valid(h):
    """
    Check if a Windows file handle is still valid.

    FIXME: temporary function

    Returns True if the handle is valid, False otherwise.
    """
    try:
        win32file.GetFileType(h)
        return True
    except TypeError:
        # TypeError happens in tests with mock.
        return False
    except pywintypes.error:
        # If GetFileType fails for any reason, the handle is likely invalid
        return False


def with_parent_lock(pathname, func, *args, **kwargs):
    """
    Run a function with a lock on the parent directory of pathname.

    This prevents race conditions where the parent directory is deleted
    while the function is running.

    If args/kwargs are provided, passes them to func.
    Otherwise, calls func() with no arguments.
    """
    if not _delete_parent_lock_needed(pathname):
        return func(*args, **kwargs)
    logger.debug('with_parent_lock(%s): acquiring lock', pathname)
    _lock_delete_parent(pathname)
    logger.debug('lock acquired: calling clean function with parent lock, is_handle_valid(%s)=%s',
                 _delete_parent_lock_key,
                 is_handle_valid(_delete_parent_lock_handle))
    try:
        return func(*args, **kwargs)
    except Exception as e:
        _close_delete_parent_lock()
        raise e from e


if IS_WINDOWS:
    atexit.register(_close_delete_parent_lock)


def delete_locked_file(pathname):
    """Delete a file that is currently in use"""
    if os.path.exists(pathname):
        MOVEFILE_DELAY_UNTIL_REBOOT = 4
        if 0 == windll.kernel32.MoveFileExW(pathname, None, MOVEFILE_DELAY_UNTIL_REBOOT):
            from ctypes import WinError
            # WinError throws the right exception based on last error.
            try:
                raise WinError()
            except PermissionError:
                # OSError has special handling in Worker.py
                # Use a special message for flagging files for later deletion
                raise OSError(
                    errno.EACCES, "Access denied in delete_locked_file()", pathname)


def delete_registry_value(key, value_name, really_delete):
    """Delete named value under the registry key.
    Return boolean indicating whether reference found and
    successful.  If really_delete is False (meaning preview),
    just check whether the value exists."""
    (hive, sub_key) = split_registry_key(key)
    try:
        if really_delete:
            hkey = winreg.OpenKey(hive, sub_key, 0, winreg.KEY_SET_VALUE)
            winreg.DeleteValue(hkey, value_name)
        else:
            hkey = winreg.OpenKey(hive, sub_key)
            winreg.QueryValueEx(hkey, value_name)
    except PermissionError:
        raise OSError(
            errno.EACCES, "Access denied in delete_registry_value()", key)
    except WindowsError as e:
        if e.winerror == errno.ENOENT:
            # ENOENT = 'file not found' means value does not exist
            return False
        raise
    return True


def delete_registry_key(parent_key, really_delete, excludekeys=None):
    """Delete registry key including any values and sub-keys.
    Return boolean whether found and success.  If really
    delete is False (meaning preview), just check whether
    the key exists."""
    parent_key = str(parent_key)  # Unicode to byte string
    excludekeys = excludekeys or []

    # Check if this key is excluded
    for exclude_path in excludekeys:
        # Normalize paths for comparison (case-insensitive)
        normalized_parent = parent_key.upper()
        normalized_exclude = exclude_path.upper()
        # Check if the key matches the exclusion (exact match or is a child of exclusion)
        if normalized_parent == normalized_exclude or \
           normalized_parent.startswith(normalized_exclude + '\\'):
            logger.debug('Skipping excluded registry key: %s', parent_key)
            return False

    (hive, parent_sub_key) = split_registry_key(parent_key)
    hkey = None
    try:
        hkey = winreg.OpenKey(hive, parent_sub_key)
    except WindowsError as e:
        if e.winerror == 2:
            # 2 = 'file not found' happens when key does not exist
            return False
    if not really_delete:
        return True
    if not hkey:
        # key not found
        return False
    keys_size = winreg.QueryInfoKey(hkey)[0]
    child_keys = [
        parent_key + '\\' + winreg.EnumKey(hkey, i) for i in range(keys_size)
    ]

    # Check if any child keys are excluded
    has_excluded_children = False
    for child_key in child_keys:
        child_deleted = delete_registry_key(child_key, True, excludekeys)
        if not child_deleted:
            has_excluded_children = True

    # If any child is excluded, preserve this parent key
    if has_excluded_children:
        logger.debug(
            'Preserving parent key with excluded children: %s', parent_key)
        return False

    try:
        winreg.DeleteKey(hive, parent_sub_key)
    except PermissionError:
        raise OSError(
            errno.EACCES, "Access denied in delete_registry_key()", parent_key)
    return True


def delete_updates():
    """Returns commands for deleting Windows Updates files

    Reference:
    https://learn.microsoft.com/en-us/troubleshoot/windows-client/installing-updates-features-roles/additional-resources-for-windows-update

    Yields commands
    """
    # Import here to avoid a circular import.
    # pylint: disable=import-outside-toplevel
    from bleachbit import Command
    if not shell.IsUserAnAdmin():
        logger.warning(
            _("Administrator privileges are required to clean Windows Updates"))
        return
    windir = os.path.expandvars('%windir%')
    dirs = glob.glob(os.path.join(windir, '$NtUninstallKB*'))
    for path_to_add in [r'%windir%\SoftwareDistribution.old',
                        r'%windir%\SoftwareDistribution.bak',
                        r'%windir%\ie7updates',
                        r'%windir%\ie8updates',
                        # see https://github.com/bleachbit/bleachbit/issues/1215 about catroot2
                        # r'%WindowsSystem%\catroot2',
                        r'%systemdrive%\windows.old',
                        r'%systemdrive%\$windows.~bt',
                        r'%systemdrive%\$windows.~ws']:
        dirs.extend(os.path.expandvars(p)
                    for p in expand_windows_system_vars(path_to_add))

    # First, delete objects that do not require services to be stopped.
    for path1 in dirs:
        for path2 in FileUtilities.children_in_directory(path1, True):
            yield Command.Delete(path2)
        if os.path.exists(path1):
            yield Command.Delete(path1)

    # Closure to bind service/start into a zero-arg callback for Command.Function
    def make_run_service(service, start):
        def run_wu_service():
            return run_net_service_command(service, start)
        return run_wu_service

    all_services = ('wuauserv', 'cryptsvc', 'bits', 'msiserver')
    restart_services = []
    for service in all_services:
        if is_service_running(service):
            restart_services.append(service)
    services_stopped = False
    sdist_dir = os.path.expandvars(r'%windir%\SoftwareDistribution')
    if not os.path.exists(sdist_dir):
        return

    for path2 in FileUtilities.children_in_directory(sdist_dir, True):
        # If we find any files, stop services.
        if not services_stopped:
            services_stopped = True
            for service in restart_services:
                # TRANSLATORS: Message in log file when stopping a Windows service.
                # The placeholder is the code name of the service.
                label = _("stop Windows service %(service)s") % {'service': service}
                yield Command.Function(None, make_run_service(service, False), label)
        yield Command.Delete(path2)
    yield Command.Delete(sdist_dir)

    if not services_stopped:
        return

    for service in restart_services:
        # TRANSLATORS: Message in log file when starting a Windows service.
        # The placeholder is the code name of the service.
        label = _("start Windows service %(service)s") % {'service': service}
        yield Command.Function(None, make_run_service(service, True), label)


def is_service_running(service):
    """Return True if service is running."""
    assert isinstance(service, str)
    service_status_code = win32serviceutil.QueryServiceStatus(service)[1]
    logger.debug('Windows service %s has current state %d',
                 service, service_status_code)
    # Throw error if status is pending.
    if service_status_code not in (1, 4, 7):
        raise RuntimeError(
            f'Unexpected service status code: {service_status_code}')
    return service_status_code == 4  # running


def run_net_service_command(service, start):
    """Start or stop a Windows service

    Args:
        service (str): Service name, e.g. 'wuauserv'.
        start (bool): True to start, False to stop.

    Behavior:
    - On success, return 0 because no space was freed.
    - Treat "already running" (start) and "not active/not started" (stop) like a success.
    - On other errors, raise RuntimeError.
    - If service has dependencies, this will stop them too.

    Reference:
    - https://github.com/bleachbit/bleachbit/issues/1932
    - https://github.com/bleachbit/bleachbit/issues/1854
    """
    assert isinstance(service, str)
    assert isinstance(start, bool)
    if start:
        action = win32serviceutil.StartService
        ignore_codes = (1056,)  # already running
        ignore_msgs = ('already',)
        verb = 'start'
        desired = win32service.SERVICE_RUNNING
        state_txt = 'RUNNING'
    else:
        action = win32serviceutil.StopServiceWithDeps
        ignore_codes = (
            1052,  # ERROR_INVALID_SERVICE_CONTROL: control not valid for this service
            1062,  # ERROR_SERVICE_NOT_ACTIVE: not active
        )
        ignore_msgs = ('not active', 'not been started', 'is not started',
                       'control is not valid')
        verb = 'stop'
        state_txt = 'STOPPED'
        desired = win32service.SERVICE_STOPPED

    try:
        action(service)
    except pywintypes.error as e:
        # Treat common benign states as success
        msg = str(e).lower()
        if getattr(e, 'winerror', None) in ignore_codes or any(s in msg for s in ignore_msgs):
            return 0
        raise RuntimeError(f'Failed to {verb} service {service}: {e}') from e

    try:
        win32serviceutil.WaitForServiceStatus(service, desired, 60)
    except Exception as wait_err:
        logger.info('WaitForServiceStatus(%s, %s, ...) had an issue: %s',
                    service, state_txt, wait_err)
    return 0


def detect_registry_key(parent_key):
    """Detect whether registry key exists"""
    try:
        parent_key = str(parent_key)  # Unicode to byte string
    except UnicodeEncodeError:
        return False
    (hive, parent_sub_key) = split_registry_key(parent_key)
    hkey = None
    try:
        hkey = winreg.OpenKey(hive, parent_sub_key)
    except WindowsError as e:
        if e.winerror == 2:
            # 2 = 'file not found' happens when key does not exist
            return False
    if not hkey:
        # key not found
        return False
    return True


def get_sid_token_48():
    """Return a 48-bit token for the current user"""
    htoken = win32security.OpenProcessToken(
        win32api.GetCurrentProcess(), win32security.TOKEN_QUERY)
    try:
        token_user = win32security.GetTokenInformation(
            htoken, win32security.TokenUser)
        sid_obj = token_user[0]
        sid_str = win32security.ConvertSidToStringSid(sid_obj)
    finally:
        win32file.CloseHandle(htoken)
    digest = hashlib.blake2b(sid_str.encode('ascii'), digest_size=6).digest()
    return base64.urlsafe_b64encode(digest).rstrip(b'=').decode('ascii')


def is_ots_elevation():
    """Return True if UAC changed credentials"""
    if os.name != 'nt':
        return False
    argv = sys.argv
    for i, arg in enumerate(argv):
        if arg == '--uac-sid-token' and i + 1 < len(argv):
            parent_token = argv[i + 1]
            try:
                return get_sid_token_48() != parent_token
            except Exception:
                return False
    return False


def elevate_privileges(uac):
    """On Windows Vista and later, try to get administrator
    privileges.  If successful, return True (so original process
    can exit).  If failed or not applicable, return False."""

    if shell.IsUserAnAdmin():
        logger.debug('already an admin (UAC not required)')
        htoken = win32security.OpenProcessToken(
            win32api.GetCurrentProcess(), win32security.TOKEN_ADJUST_PRIVILEGES | win32security.TOKEN_QUERY)
        newPrivileges = [
            (win32security.LookupPrivilegeValue(None, "SeBackupPrivilege"),
             win32security.SE_PRIVILEGE_ENABLED),
            (win32security.LookupPrivilegeValue(None, "SeRestorePrivilege"),
             win32security.SE_PRIVILEGE_ENABLED),
        ]
        win32security.AdjustTokenPrivileges(htoken, 0, newPrivileges)
        win32file.CloseHandle(htoken)
        return False
    elif not uac:
        return False

    if hasattr(sys, 'frozen'):
        # running frozen in py2exe
        exe = sys.executable
        parameters = "--gui --no-uac"
    else:
        pyfile = os.path.join(bleachbit.bleachbit_exe_path, 'bleachbit.py')
        # If the Python file is on a network drive, do not offer the UAC because
        # the administrator may not have privileges and user will not be
        # prompted.
        if len(pyfile) > 0 and path_on_network(pyfile):
            logger.debug(
                "debug: skipping UAC because '%s' is on network", pyfile)
            return False
        parameters = '"%s" --gui --no-uac' % pyfile
        exe = sys.executable

    try:
        token = get_sid_token_48()
        parameters = f"{parameters} --uac-sid-token {token}"
    except Exception as e:
        logger.error('could not compute SID token: %s', e)

    parameters = _add_command_line_parameters(parameters)

    logger.debug('elevate_privileges() exe=%s, parameters=%s', exe, parameters)

    rc = None
    try:
        rc = shell.ShellExecuteEx(lpVerb='runas',
                                  lpFile=exe,
                                  lpParameters=parameters,
                                  nShow=win32con.SW_SHOW)
    except pywintypes.error as e:
        if 1223 == e.winerror:
            logger.debug('user denied the UAC dialog')
            return False
        raise

    logger.debug('ShellExecuteEx=%s', rc)

    if isinstance(rc, dict):
        return True

    return False


def _add_command_line_parameters(parameters):
    """
    Add any command line parameters such as --debug-log.
    """
    if '--context-menu' in sys.argv:
        return '{} {} "{}"'.format(parameters, ' '.join(sys.argv[1:-1]), sys.argv[-1])

    return '{} {}'.format(parameters, ' '.join(sys.argv[1:]))


def empty_recycle_bin(path, really_delete):
    """Empty the recycle bin or preview its size.

    If the recycle bin is empty, it is not emptied again to avoid an error.

    Keyword arguments:
    path          -- A drive, folder or None.  None refers to all recycle bins.
    really_delete -- If True, then delete.  If False, then just preview.

    Returns:
    The number of bytes that were freed or would be freed.
    """
    (recycle_bin_size, num_files) = shell.SHQueryRecycleBin(path)
    if really_delete and num_files > 0:
        # Trying to delete an empty Recycle Bin on Vista/7 causes a
        # 'catastrophic failure'
        flags = shellcon.SHERB_NOSOUND | shellcon.SHERB_NOCONFIRMATION | shellcon.SHERB_NOPROGRESSUI
        try:
            shell.SHEmptyRecycleBin(None, path, flags)
        except Exception:
            logger.exception(
                'error in SHEmptyRecycleBin(): recycle_bin_size=%s, num_files=%r, path=%r',
                FileUtilities.bytes_to_human(recycle_bin_size), num_files, path)
            raise
    return recycle_bin_size


def flush_dns():
    """Flush the DNS resolver cache

    Returns 0 on success.
    Raises RuntimeError on failure.
    """
    args = ['ipconfig', '/flushdns']
    (rc, stdout, stderr) = General.run_external(args)
    if 0 != rc:
        raise RuntimeError(
            f'Command: {args}\nReturn code: {rc}\nStdout: {stdout}\nStderr: {stderr}')
    return 0


def clear_clipboard():
    """Clear the clipboard"""
    _open_clipboard()
    try:
        win32clipboard.EmptyClipboard()
    except Exception:
        logger.exception('error clearing clipboard')
    finally:
        win32clipboard.CloseClipboard()


def get_clipboard_paths():
    """Return a tuple of Unicode pathnames from the clipboard"""
    _open_clipboard()
    path_list = ()
    try:
        path_list = win32clipboard.GetClipboardData(win32clipboard.CF_HDROP)
    except TypeError:
        pass
    finally:
        win32clipboard.CloseClipboard()
    return path_list


def _open_clipboard(timeout=1.0):
    """Open the clipboard with retry logic"""
    deadline = time.time() + timeout
    while True:
        try:
            return win32clipboard.OpenClipboard()
        except pywintypes.error as e:
            winerror = getattr(e, 'winerror', e.args[0] if e.args else None)
            if winerror != 5 or time.time() >= deadline:
                raise
            time.sleep(0.05)


def get_fixed_drives():
    """Yield each fixed drive"""
    for drive in win32api.GetLogicalDriveStrings().split('\x00'):
        if win32file.GetDriveType(drive) == win32file.DRIVE_FIXED:
            # Microsoft Office 2010 Starter creates a virtual drive that
            # looks much like a fixed disk but isdir() returns false
            # and free_space() returns access denied.
            # https://bugs.launchpad.net/bleachbit/+bug/1474848
            if os.path.isdir(drive):
                yield drive


def get_known_folder_path(folder_name):
    """Return the path of a folder by its Folder ID

    Requires Windows Vista, Server 2008, or later

    Based on the code Michael Kropat (mkropat) from
    <https://gist.github.com/mkropat/7550097>
    licensed  under the GNU GPL"""

    class GUID(ctypes.Structure):
        _fields_ = [
            ("Data1", wintypes.DWORD),
            ("Data2", wintypes.WORD),
            ("Data3", wintypes.WORD),
            ("Data4", wintypes.BYTE * 8)
        ]

        def __init__(self, uuid_):
            ctypes.Structure.__init__(self)
            self.Data1, self.Data2, self.Data3, self.Data4[
                0], self.Data4[1], rest = uuid_.fields
            for i in range(2, 8):
                self.Data4[i] = rest >> (8 - i - 1) * 8 & 0xff

    class FOLDERID:
        LocalAppDataLow = UUID(
            '{A520A1A4-1780-4FF6-BD18-167343C5AF16}')
        Fonts = UUID('{FD228CB7-AE11-4AE3-864C-16F3910AB8FE}')

    class UserHandle:
        current = wintypes.HANDLE(0)

    _CoTaskMemFree = windll.ole32.CoTaskMemFree
    _CoTaskMemFree.restype = None
    _CoTaskMemFree.argtypes = [ctypes.c_void_p]

    try:
        _SHGetKnownFolderPath = windll.shell32.SHGetKnownFolderPath
    except AttributeError:
        # Not supported on Windows XP
        return None
    _SHGetKnownFolderPath.argtypes = [
        ctypes.POINTER(GUID), wintypes.DWORD, wintypes.HANDLE, ctypes.POINTER(
            ctypes.c_wchar_p)
    ]

    class PathNotFoundException(Exception):
        pass

    folderid = getattr(FOLDERID, folder_name)
    fid = GUID(folderid)
    pPath = ctypes.c_wchar_p()
    S_OK = 0
    if _SHGetKnownFolderPath(ctypes.byref(fid), 0, UserHandle.current, ctypes.byref(pPath)) != S_OK:
        raise PathNotFoundException(folder_name)
    path = pPath.value
    _CoTaskMemFree(pPath)
    return path


def get_recycle_bin():
    """Yield a list of files in the recycle bin"""
    pidl = shell.SHGetSpecialFolderLocation(0, shellcon.CSIDL_BITBUCKET)
    desktop = shell.SHGetDesktopFolder()
    h = desktop.BindToObject(pidl, None, shell.IID_IShellFolder)
    for item in h:
        path = h.GetDisplayNameOf(item, shellcon.SHGDN_FORPARSING)
        if FileUtilities.is_normal_directory(path):
            # Return the contents of a normal directory, but do
            # not recurse Windows symlinks in the Recycle Bin.
            yield from FileUtilities.children_in_directory(path, True)
        yield path


def get_windows_version():
    """Get the Windows major and minor version in a decimal like 10.0"""
    v = win32api.GetVersionEx(0)
    vstr = '%d.%d' % (v[0], v[1])
    return Decimal(vstr)


def is_junction(path):
    """Check whether the path is a junction (mount point) on Windows.

    Python 3.12 added os.is_junction()
    https://docs.python.org/3/library/os.html#os.DirEntry.is_junction
    """
    if sys.platform != 'win32':
        return False
    if hasattr(os, 'is_junction'):
        return os.is_junction(path)

    # Get reparse tag from stat result
    try:
        stat_result = os.stat(path, follow_symlinks=False)
        tag = getattr(stat_result, 'st_reparse_tag', None)
        if tag is not None:
            return tag == IO_REPARSE_TAG_MOUNT_POINT
    except (OSError, AttributeError):
        pass

    attr = windll.kernel32.GetFileAttributesW(path)
    # INVALID_FILE_ATTRIBUTES (0xFFFFFFFF) indicates GetFileAttributesW failed
    # On 64-bit Python, ctypes may interpret this as signed -1 instead of unsigned 0xFFFFFFFF
    if attr == 0xFFFFFFFF or attr == -1:
        error_code = windll.kernel32.GetLastError()
        logger.error(
            'GetFileAttributesW() failed for path %s with error code %d', path, error_code)
        return False
    return bool(attr & FILE_ATTRIBUTE_REPARSE_POINT)


def load_i18n_dll():
    """Load internationalization library

    BleachBit 4.6.2 with Python 3.4 and GTK 3.18 had libintl-8.dll.

    BleachBit 5.0.0 with Python 3.10 and GTK 3.24 built on vcpkg
    has intl-8.dll.

    Either way, it comes from gettext.

    Returns None if the dll is not available.
    """
    dirs = set([bleachbit.bleachbit_exe_path, os.path.dirname(sys.executable)])
    lib_path = None
    for dir in dirs:
        lib_path = os.path.join(dir, 'intl-8.dll')
        if os.path.exists(lib_path):
            break
    if not lib_path:
        logger.warning(
            'internationalization library was not found, so translations will not work.')
        return
    try:
        libintl = ctypes.cdll.LoadLibrary(lib_path)
    except Exception as e:
        logger.warning('error in LoadLibrary(%s): %s', lib_path, e)
        return

    # Configure DLL function prototypes
    libintl.bindtextdomain.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
    libintl.bindtextdomain.restype = ctypes.c_char_p
    libintl.bind_textdomain_codeset.argtypes = [
        ctypes.c_char_p, ctypes.c_char_p]
    libintl.textdomain.argtypes = [ctypes.c_char_p]

    if hasattr(libintl, "libintl_wbindtextdomain"):
        libintl.libintl_wbindtextdomain.argtypes = [
            ctypes.c_char_p, ctypes.c_wchar_p]
        libintl.libintl_wbindtextdomain.restype = ctypes.c_wchar_p

    return libintl


def move_to_recycle_bin(path):
    """Move 'path' into recycle bin"""
    shell.SHFileOperation(
        (0, shellcon.FO_DELETE, path, None, shellcon.FOF_ALLOWUNDO | shellcon.FOF_NOCONFIRMATION))


def parse_windows_build(build=None):
    """
    Parse build string like 1.2.3 or 1.2 to numeric,
    ignoring the third part, if present.
    """
    if not build:
        # If not given, default to current system's version
        return get_windows_version()
    return Decimal('.'.join(build.split('.')[0:2]))


def path_on_network(path):
    """Check whether 'path' is on a network drive"""
    drive = os.path.splitdrive(path)[0]
    if drive.startswith(r'\\'):
        return True
    return win32file.GetDriveType(drive) == win32file.DRIVE_REMOTE


def shell_change_notify():
    """Notify the Windows shell of update.

    Used in windows_explorer.xml."""
    shell.SHChangeNotify(shellcon.SHCNE_ASSOCCHANGED, shellcon.SHCNF_IDLIST,
                         None, None)
    return 0


def set_environ(varname, path):
    """Define an environment variable for use in CleanerML and Winapp2.ini"""
    if not path:
        return
    if varname in os.environ:
        # logger.debug('set_environ(%s, %s): skipping because environment variable is already defined', varname, path)
        if 'nt' == os.name:
            os.environ[varname] = os.path.expandvars('%%%s%%' % varname)
        # Do not redefine the environment variable when it already exists
        # But re-encode them with utf-8 instead of mbcs
        return
    try:
        if not os.path.exists(path):
            raise RuntimeError(
                'Variable %s points to a non-existent path %s' % (varname, path))
        os.environ[varname] = path
    except:
        logger.exception(
            'set_environ(%s, %s): exception when setting environment variable', varname, path)


def setup_environment():
    """Define any extra environment variables"""

    # These variables are for use in CleanerML and Winapp2.ini.
    csidl_to_environ('commonappdata', shellcon.CSIDL_COMMON_APPDATA)
    csidl_to_environ('documents', shellcon.CSIDL_PERSONAL)
    # Windows XP does not define localappdata, but Windows Vista and 7 do
    csidl_to_environ('localappdata', shellcon.CSIDL_LOCAL_APPDATA)
    csidl_to_environ('music', shellcon.CSIDL_MYMUSIC)
    csidl_to_environ('pictures', shellcon.CSIDL_MYPICTURES)
    csidl_to_environ('video', shellcon.CSIDL_MYVIDEO)
    # LocalLowAppData does not have a CSIDL for use with
    # SHGetSpecialFolderPath. Instead, it is identified using
    # SHGetKnownFolderPath in Windows Vista and later
    try:
        path = get_known_folder_path('LocalAppDataLow')
    except:
        logger.exception('exception identifying LocalAppDataLow')
    else:
        set_environ('LocalAppDataLow', path)
    # %cd% can be helpful for cleaning portable applications when
    # BleachBit is portable. It is the same variable name as defined by
    # cmd.exe .
    set_environ('cd', os.getcwd())

    # XDG_DATA_DIRS environment variable needs to be set with both GTK icons and
    # `glib-2.0\schemas\gschemas.compiled`.
    # The latter is required by the make chaff dialog.
    # https://github.com/bleachbit/bleachbit/issues/1444
    # https://github.com/bleachbit/bleachbit/issues/1780
    if os.environ.get('XDG_DATA_DIRS'):
        return
    xdg_data_dirs = [os.path.dirname(sys.executable) +
                     '\\share',
                     os.getcwd() + '\\share']

    found_dir = False
    for xdg_data_dir in xdg_data_dirs:
        xdg_data_fn = os.path.join(
            xdg_data_dir, 'glib-2.0', 'schemas', 'gschemas.compiled')
        if os.path.exists(xdg_data_fn):
            found_dir = True
            break
    if found_dir:
        logger.debug('XDG_DATA_DIRS=%s', xdg_data_dir)
        set_environ('XDG_DATA_DIRS', xdg_data_dir)
    else:
        logger.warning('XDG_DATA_DIRS not set and gschemas.compiled not found')


def split_registry_key(full_key):
    r"""Given a key like HKLM\Software split into tuple (hive, key).
    Used internally."""
    assert len(full_key) >= 6
    [k1, k2] = full_key.split("\\", 1)
    hive_map = {
        'HKCR': winreg.HKEY_CLASSES_ROOT,
        'HKCU': winreg.HKEY_CURRENT_USER,
        'HKLM': winreg.HKEY_LOCAL_MACHINE,
        'HKU': winreg.HKEY_USERS}
    if k1 not in hive_map:
        raise RuntimeError("Invalid Windows registry hive '%s'" % k1)
    return hive_map[k1], k2


def read_registry_key(full_key, value_name):
    try:
        (hive, sub_key) = split_registry_key(full_key)
    except RuntimeError as e:
        return None
    try:
        with winreg.OpenKey(hive, sub_key, 0, winreg.KEY_QUERY_VALUE) as hkey:
            (reg_value, reg_type) = winreg.QueryValueEx(hkey, value_name)
            if reg_type == winreg.REG_EXPAND_SZ or reg_type == winreg.REG_SZ:
                return reg_value
            else:
                return None
    except OSError as e:
        if e.winerror == errno.ENOENT:
            # ENOENT = 'file not found' means value does not exist
            return None
        raise


def symlink_or_copy(src, dst):
    """Symlink with fallback to copy

    Symlink is faster and uses virtually no storage, but it it requires administrator
    privileges or Windows developer mode.

    If symlink is not available, just copy the file.
    """
    try:
        os.symlink(src, dst)
        logger.debug('linked %s to %s', src, dst)
    except (PermissionError, OSError):
        shutil.copy(src, dst)
        logger.debug('copied %s to %s', src, dst)


def has_fontconfig_cache(font_conf_file):
    dom = xml.dom.minidom.parse(font_conf_file)
    fc_element = dom.getElementsByTagName('fontconfig')[0]
    cachefile = 'd031bbba323fd9e5b47e0ee5a0353f11-le32d8.cache-6'
    expanded_localdata = os.path.expandvars('%LOCALAPPDATA%')
    expanded_homepath = os.path.join(os.path.expandvars(
        '%HOMEDRIVE%'), os.path.expandvars('%HOMEPATH%'))
    for dir_element in fc_element.getElementsByTagName('cachedir'):

        if dir_element.firstChild.nodeValue == 'LOCAL_APPDATA_FONTCONFIG_CACHE':
            dirpath = os.path.join(expanded_localdata, 'fontconfig', 'cache')
        elif dir_element.firstChild.nodeValue == 'fontconfig' and dir_element.getAttribute('prefix') == 'xdg':
            dirpath = os.path.join(expanded_homepath, '.cache', 'fontconfig')
        elif dir_element.firstChild.nodeValue == '~/.fontconfig':
            dirpath = os.path.join(expanded_homepath, '.fontconfig')
        else:
            # user has entered a custom directory
            dirpath = dir_element.firstChild.nodeValue

        if dirpath and os.path.exists(os.path.join(dirpath, cachefile)):
            return True

    return False


def get_font_conf_file():
    """Return the full path to fonts.conf"""
    if hasattr(sys, 'frozen'):
        # running inside py2exe
        return os.path.join(bleachbit.bleachbit_exe_path, 'etc', 'fonts', 'fonts.conf')

    import gi
    gnome_dir = os.path.join(os.path.dirname(
        os.path.dirname(gi.__file__)), 'gnome')
    if not os.path.isdir(gnome_dir):
        # BleachBit is running from a stand-alone Python installation.
        gnome_dir = os.path.join(sys.exec_prefix, '..', '..')
    return os.path.join(gnome_dir, 'etc', 'fonts', 'fonts.conf')


class SplashThread(Thread):
    _class_atom = None

    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs={}, Verbose=None):
        super().__init__(group, self._show_splash_screen, name, args, kwargs)
        self.daemon = True
        self._splash_screen_started = Event()
        self._splash_screen_closed = Event()
        self._splash_screen_handle = None
        self._splash_screen_height = None
        self._splash_screen_width = None
        self._startup_error = None
        self._thread_id = None

    def start(self):
        if self.ident is not None:
            logger.debug('SplashThread was already started')
            return
        try:
            Thread.start(self)
        except RuntimeError:
            logger.debug('SplashThread could not be started', exc_info=True)
            return
        started = self._splash_screen_started.wait(timeout=10)
        if not started:
            logger.warning('SplashThread did not start within timeout')
        else:
            logger.debug('SplashThread started')

        if self._startup_error:
            logger.debug('Splash screen disabled due to startup error')

    def run(self):
        try:
            self._thread_id = win32api.GetCurrentThreadId()
            self._splash_screen_handle = self._show_splash_screen()
        except Exception as exc:
            self._startup_error = exc
            logger.exception('SplashThread failed to initialize splash screen')
        finally:
            self._splash_screen_started.set()

        if self._startup_error:
            self._splash_screen_closed.set()
            return

        # Dispatch messages
        try:
            win32gui.PumpMessages()
        except Exception:
            logger.exception('SplashThread message pump failed')
        finally:
            self._splash_screen_handle = None
            self._splash_screen_closed.set()

    def join(self, timeout=None):
        self.close(timeout)

    def close(self, timeout=None):
        if not self.is_alive():
            return
        splash_delay = get_splash_screen_delay_seconds()
        if splash_delay > 0:
            logger.debug(
                'Delaying splash screen close by %s seconds', splash_delay)
            time.sleep(splash_delay)
        self._request_close()
        Thread.join(self, timeout=timeout)

    def _request_close(self):
        """Ask the splash window to close without risking a GUI crash."""
        if not self._splash_screen_handle:
            self._post_quit_message()
            return

        hWindow = self._splash_screen_handle
        try:
            if not win32gui.IsWindow(hWindow):
                self._splash_screen_handle = None
                self._post_quit_message()
                return
        except Exception:
            logger.debug('Could not verify splash screen window', exc_info=True)

        self._hide_window(hWindow)
        if self._send_close_message(hWindow):
            return

        try:
            win32gui.PostMessage(hWindow, win32con.WM_CLOSE, 0, 0)
        except Exception:
            logger.debug('Failed to post close message to splash screen',
                         exc_info=True)
            self._post_quit_message()

    def _hide_window(self, hWindow):
        """Hide the splash window before closing it."""
        try:
            show_window_async = ctypes.windll.user32.ShowWindowAsync
            show_window_async.argtypes = [wintypes.HWND, ctypes.c_int]
            show_window_async.restype = wintypes.BOOL
            show_window_async(wintypes.HWND(hWindow), win32con.SW_HIDE)
        except Exception:
            logger.debug('Failed to hide splash screen', exc_info=True)

    def _send_close_message(self, hWindow):
        """Synchronously close the splash window with a timeout."""
        flags = getattr(win32con, 'SMTO_ABORTIFHUNG', 0x0002)
        try:
            send_message_timeout_w = ctypes.windll.user32.SendMessageTimeoutW
            send_message_timeout_w.argtypes = [
                wintypes.HWND,
                wintypes.UINT,
                wintypes.WPARAM,
                wintypes.LPARAM,
                wintypes.UINT,
                wintypes.UINT,
                ctypes.POINTER(ctypes.c_void_p),
            ]
            send_message_timeout_w.restype = wintypes.LPARAM
            message_result = ctypes.c_void_p()
            result = send_message_timeout_w(
                wintypes.HWND(hWindow), win32con.WM_CLOSE, 0, 0, flags,
                SPLASH_CLOSE_TIMEOUT_MS, ctypes.byref(message_result))
            if result:
                return True
            logger.debug('SendMessageTimeoutW failed: %s',
                         ctypes.get_last_error())
        except Exception:
            logger.debug('Failed to send close message with ctypes',
                         exc_info=True)

        return False

    def _post_quit_message(self):
        """Exit the splash message pump when the window cannot be addressed."""
        if self._thread_id is None:
            return
        post_thread_message = getattr(win32api, 'PostThreadMessage', None)
        if post_thread_message is None:
            return
        try:
            post_thread_message(self._thread_id, win32con.WM_QUIT, 0, 0)
        except Exception:
            logger.debug('Failed to post quit message to splash thread',
                         exc_info=True)

    def get_icon_path(self):
        """Return the full path to icon file"""
        if hasattr(sys, 'frozen'):
            # running frozen in py2exe
            icon_dir = Path(sys.argv[0]).parent / 'share'
        else:
            # running from source, and `__file__`` may be at either level
            # - `tests/TestWindows.py`
            # - `bleachbit.py``
            module_dir = Path(__file__).absolute().parent
            icon_dir = module_dir / 'windows'
            icon_path = module_dir / 'bleachbit.ico'
            if not icon_path.exists():
                icon_dir = module_dir.parent / 'windows'
        return (icon_dir / 'bleachbit.ico').absolute()

    def calculate_window_position(self, display_width, display_height):
        """Calculate centered window position and size"""
        x = (display_width - SPLASH_ICON_SIZE_PX) // 2
        y = (display_height - SPLASH_ICON_SIZE_PX) // 2
        return (x, y, SPLASH_ICON_SIZE_PX, SPLASH_ICON_SIZE_PX)

    def _render_splash(self, hWnd):
        """Load image and render the splash screen"""
        filename = self.get_icon_path()
        if not filename.exists():
            logger.warning(
                'Icon file not found: %s with current working directory: %s.', filename, os.getcwd())
            return

        user32 = ctypes.windll.user32
        gdi32 = ctypes.windll.gdi32
        icon_size = 256
        flags = win32con.LR_LOADFROMFILE
        hIcon = win32gui.LoadImage(
            0, str(filename), win32con.IMAGE_ICON, icon_size, icon_size, flags)
        if not hIcon:
            return

        user32.UpdateLayeredWindow.argtypes = [
            wintypes.HWND,
            wintypes.HDC,
            ctypes.POINTER(_POINT),
            ctypes.POINTER(_SIZE),
            wintypes.HDC,
            ctypes.POINTER(_POINT),
            wintypes.COLORREF,
            ctypes.POINTER(_BLENDFUNCTION),
            wintypes.DWORD,
        ]
        user32.UpdateLayeredWindow.restype = wintypes.BOOL
        gdi32.CreateCompatibleDC.argtypes = [wintypes.HDC]
        gdi32.CreateCompatibleDC.restype = wintypes.HDC
        gdi32.CreateDIBSection.argtypes = [
            wintypes.HDC,
            ctypes.POINTER(_BITMAPINFO),
            wintypes.UINT,
            ctypes.POINTER(ctypes.c_void_p),
            wintypes.HANDLE,
            wintypes.DWORD,
        ]
        gdi32.CreateDIBSection.restype = wintypes.HBITMAP
        gdi32.SelectObject.argtypes = [wintypes.HDC, wintypes.HGDIOBJ]
        gdi32.SelectObject.restype = wintypes.HGDIOBJ
        gdi32.DeleteObject.argtypes = [wintypes.HGDIOBJ]
        gdi32.DeleteObject.restype = wintypes.BOOL
        gdi32.DeleteDC.argtypes = [wintypes.HDC]
        gdi32.DeleteDC.restype = wintypes.BOOL

        hdc_screen = None
        hdc_mem = None
        hbm = None
        hold = None
        try:
            hdc_screen = win32gui.GetDC(0)
            if not hdc_screen:
                return

            hdc_mem = gdi32.CreateCompatibleDC(hdc_screen)
            if not hdc_mem:
                return

            bitmap_info = _BITMAPINFO()
            bitmap_info.bmiHeader.biSize = ctypes.sizeof(_BITMAPINFOHEADER)
            bitmap_info.bmiHeader.biWidth = icon_size
            bitmap_info.bmiHeader.biHeight = -icon_size
            bitmap_info.bmiHeader.biPlanes = 1
            bitmap_info.bmiHeader.biBitCount = 32
            bitmap_info.bmiHeader.biCompression = _BI_RGB

            bits = ctypes.c_void_p()
            hbm = gdi32.CreateDIBSection(
                hdc_screen,
                ctypes.byref(bitmap_info),
                _DIB_RGB_COLORS,
                ctypes.byref(bits),
                None,
                0)
            if not hbm or not bits.value:
                return

            hold = gdi32.SelectObject(hdc_mem, hbm)
            ctypes.memset(bits.value, 0, icon_size * icon_size * 4)

            try:
                win32gui.DrawIconEx(
                    hdc_mem, 0, 0, hIcon, icon_size, icon_size, 0, 0, win32con.DI_NORMAL)
                rect = win32gui.GetWindowRect(hWnd)
                pt_dst = _POINT(rect[0], rect[1])
                pt_src = _POINT(0, 0)
                size_wnd = _SIZE(icon_size, icon_size)
                blend = _BLENDFUNCTION(_AC_SRC_OVER, 0, 255, _AC_SRC_ALPHA)
                success = user32.UpdateLayeredWindow(
                    hWnd,
                    hdc_screen,
                    ctypes.byref(pt_dst),
                    ctypes.byref(size_wnd),
                    hdc_mem,
                    ctypes.byref(pt_src),
                    0,
                    ctypes.byref(blend),
                    _ULW_ALPHA)
                if not success:
                    logger.warning('UpdateLayeredWindow failed: %s',
                                   ctypes.get_last_error())
            finally:
                if hold:
                    gdi32.SelectObject(hdc_mem, hold)
        finally:
            if hbm:
                gdi32.DeleteObject(hbm)
            if hdc_mem:
                gdi32.DeleteDC(hdc_mem)
            if hdc_screen:
                win32gui.ReleaseDC(0, hdc_screen)
            try:
                win32gui.DestroyIcon(hIcon)
            except Exception:
                pass

    def _safe_render_splash(self, hWnd):
        """Render the splash screen without letting paint failures escape."""
        try:
            self._render_splash(hWnd)
        except Exception:
            logger.exception('Failed to render splash screen')

    def _register_window_class(self, wndClass):
        """Register splash screen window class, handling reuse."""
        cached_atom = self.__class__._class_atom
        if cached_atom:
            return cached_atom

        try:
            atom = win32gui.RegisterClass(wndClass)
        except pywintypes.error as err:
            if getattr(err, 'winerror', None) != 1410:  # ERROR_CLASS_ALREADY_EXISTS
                raise
            # Try to get class info if function exists
            try:
                existing = win32gui.GetClassInfo(
                    wndClass.hInstance, wndClass.lpszClassName)
                if isinstance(existing, tuple):
                    atom = existing[0]
                else:
                    atom = existing
                if not atom:
                    raise
            except AttributeError:
                # GetClassInfo doesn't exist, use a fallback atom value
                atom = 1234  # Fallback atom for testing/compatibility
            except Exception:
                # GetClassInfo failed, use fallback
                atom = 1234

        self.__class__._class_atom = atom
        return atom

    def _show_splash_screen(self):
        """Create window and render the splash screen."""
        # get instance handle
        hInstance = win32api.GetModuleHandle()

        # the class name
        className = 'SimpleWin32'

        # create and initialize window class
        wndClass = win32gui.WNDCLASS()
        wndClass.style = win32con.CS_HREDRAW | win32con.CS_VREDRAW
        wndClass.lpfnWndProc = self.wndProc
        wndClass.hInstance = hInstance
        wndClass.hIcon = win32gui.LoadIcon(0, win32con.IDI_APPLICATION)
        wndClass.hCursor = win32gui.LoadCursor(0, win32con.IDC_ARROW)
        wndClass.hbrBackground = win32gui.GetStockObject(win32con.NULL_BRUSH)
        wndClass.lpszClassName = className

        # register window class
        wndClassAtom = self._register_window_class(wndClass)

        displayWidth = win32api.GetSystemMetrics(0)
        displayHeight = win32api.GetSystemMetrics(1)
        windowPosX, windowPosY, self._splash_screen_width, self._splash_screen_height = self.calculate_window_position(
            displayWidth, displayHeight)

        hWindow = None
        try:
            hWindow = win32gui.CreateWindowEx(
                win32con.WS_EX_LAYERED | win32con.WS_EX_TOOLWINDOW | win32con.WS_EX_TOPMOST,
                wndClassAtom,  # it seems message dispatching only works with the atom, not the class name
                'Bleachbit splash screen',
                win32con.WS_POPUP |
                win32con.WS_VISIBLE,
                windowPosX,
                windowPosY,
                self._splash_screen_width,
                self._splash_screen_height,
                0,
                0,
                hInstance,
                None)
            if not hWindow:
                raise RuntimeError('CreateWindowEx returned no splash window')
            self._splash_screen_handle = hWindow
        except Exception:
            if hWindow:
                try:
                    win32gui.DestroyWindow(hWindow)
                except Exception:
                    logger.debug('Failed to destroy incomplete splash window',
                                 exc_info=True)
            raise

        try:
            win32gui.UpdateWindow(hWindow)
        except Exception:
            logger.debug('Failed to update splash screen window',
                         exc_info=True)
        self._safe_render_splash(hWindow)

        try:
            is_splash_screen_on_top = self._force_set_foreground_window(
                hWindow)
        except Exception:
            logger.debug('Failed to foreground splash screen', exc_info=True)
            is_splash_screen_on_top = False
        logger.debug(
            'Is splash screen on top: {}'.format(is_splash_screen_on_top)
        )

        return hWindow

    def _force_set_foreground_window(self, hWindow):
        """Keep the splash screen on top"""
        # As there are some restrictions about which processes can call SetForegroundWindow as described here:
        # https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-setforegroundwindow
        # we try consecutively three different ways to show the splash screen on top of all other windows.

        # Solution 1: Pressing alt key unlocks SetForegroundWindow
        # https://stackoverflow.com/questions/14295337/win32gui-setactivewindow-error-the-specified-procedure-could-not-be-found
        # Not using win32com.client.Dispatch like in the link because there are problems when building with py2exe.
        ALT_KEY = win32con.VK_MENU
        RIGHT_ALT = 0xb8
        win32api.keybd_event(ALT_KEY, RIGHT_ALT, 0, 0)
        win32api.keybd_event(ALT_KEY, RIGHT_ALT, win32con.KEYEVENTF_KEYUP, 0)
        win32gui.ShowWindow(hWindow, win32con.SW_SHOW)
        try:
            win32gui.SetForegroundWindow(hWindow)
        except Exception as e:
            exc_message = str(e)
            logger.debug(
                'Failed attempt to show splash screen with keybd_event: {}'.format(
                    exc_message)
            )

        if win32gui.GetForegroundWindow() == hWindow:
            return True

        # Solution 2: Attaching current thread to the foreground thread in order to use BringWindowToTop
        # https://shlomio.wordpress.com/2012/09/04/solved-setforegroundwindow-win32-api-not-always-works/
        foreground_thread_id, _foreground_process_id = win32process.GetWindowThreadProcessId(
            win32gui.GetForegroundWindow())
        appThread = win32api.GetCurrentThreadId()

        if foreground_thread_id != appThread:

            try:
                win32process.AttachThreadInput(
                    foreground_thread_id, appThread, True)
                win32gui.BringWindowToTop(hWindow)
                win32gui.ShowWindow(hWindow, win32con.SW_SHOW)
                win32process.AttachThreadInput(
                    foreground_thread_id, appThread, False)
            except Exception as e:
                exc_message = str(e)
                logger.debug(
                    'Failed attempt to show splash screen with AttachThreadInput: {}'.format(
                        exc_message)
                )

        else:
            win32gui.BringWindowToTop(hWindow)
            win32gui.ShowWindow(hWindow, win32con.SW_SHOW)

        if win32gui.GetForegroundWindow() == hWindow:
            return True

        # Solution 3: Working with timers that lock/unlock SetForegroundWindow
        # https://gist.github.com/EBNull/1419093
        try:
            timeout = win32gui.SystemParametersInfo(
                win32con.SPI_GETFOREGROUNDLOCKTIMEOUT)
            win32gui.SystemParametersInfo(
                win32con.SPI_SETFOREGROUNDLOCKTIMEOUT, 0, win32con.SPIF_SENDCHANGE)
            win32gui.BringWindowToTop(hWindow)
            win32gui.SetForegroundWindow(hWindow)
            win32gui.SystemParametersInfo(
                win32con.SPI_SETFOREGROUNDLOCKTIMEOUT, timeout, win32con.SPIF_SENDCHANGE)
        except Exception as e:
            exc_message = str(e)
            logger.debug(
                'Failed attempt to show splash screen with SystemParametersInfo: {}'.format(
                    exc_message)
            )

        if win32gui.GetForegroundWindow() == hWindow:
            return True

        # Solution 4: If on some machines the splash screen still doesn't come on top, we can try
        # the following solution that combines attaching to a thread and timers:
        # https://www.codeproject.com/Tips/76427/How-to-Bring-Window-to-Top-with-SetForegroundWindo

        return False

    def wndProc(self, hWnd, message, wParam, lParam):
        """Window procedure for handling messages"""

        if message == win32con.WM_CREATE:
            self._safe_render_splash(hWnd)
            return 0

        if message == win32con.WM_ERASEBKGND:
            return 1

        if message == win32con.WM_PAINT:
            hDC, paintStruct = win32gui.BeginPaint(hWnd)
            try:
                self._safe_render_splash(hWnd)
            finally:
                win32gui.EndPaint(hWnd, paintStruct)
            return 0

        elif message == win32con.WM_CLOSE:
            try:
                win32gui.DestroyWindow(hWnd)
                return 0
            except Exception:
                logger.debug('Failed to destroy splash screen window',
                             exc_info=True)
                return win32gui.DefWindowProc(hWnd, message, wParam, lParam)

        elif message == win32con.WM_DESTROY:
            if hWnd == self._splash_screen_handle:
                self._splash_screen_handle = None
            self._splash_screen_closed.set()
            win32gui.PostQuitMessage(0)
            return 0

        else:
            return win32gui.DefWindowProc(hWnd, message, wParam, lParam)


splash_thread = SplashThread()
atexit.register(splash_thread.close)



# vim: ts=4:sw=4:expandtab

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Dialog used to verify text rendering on Windows.

This is a temporary workaround.
"""


from bleachbit.GtkShim import require_gtk, Gtk
from bleachbit.Language import get_text as _

require_gtk()

from gi.repository import Pango


ALT_FONT = "Arial 14"

RESPONSE_TEXT_BLURRY = 1001
RESPONSE_TEXT_UNREADABLE = 1002


def _apply_alt_font(widget):
    """Apply the alternate font to a label or label-like widget."""
    desc = Pango.FontDescription(ALT_FONT)
    widget.modify_font(desc)


def create_font_check_dialog(parent=None):
    """Return the dialog prompting the user about font rendering."""
    dialog = Gtk.Dialog(
        # TRANSLATORS: Title of the font check dialog
        title=_("Font check"),
        transient_for=parent,
        modal=True
    )
    dialog.set_border_width(12)
    dialog.set_resizable(False)

    dialog.add_buttons(
        # TRANSLATORS: Button label in the font check dialog
        _("Text is okay"), Gtk.ResponseType.YES,
        # TRANSLATORS: Button label in the font check dialog. This is one of
        # three buttons the user can click to report how the text appears.
        # The user clicks this if the text is readable but appears fuzzy.
        _("Text is blurry"), RESPONSE_TEXT_BLURRY,
        # TRANSLATORS: Button label in the font check dialog in case the font
        # looks like an alien language. This is one of three buttons the user
        # can click to report how the text appears.
        _("Text is unrecognizable"), RESPONSE_TEXT_UNREADABLE,
    )

    # To ensure user can read them, apply the alternate font to all buttons.
    for response_id in (
        Gtk.ResponseType.YES,
        RESPONSE_TEXT_BLURRY,
        RESPONSE_TEXT_UNREADABLE,
    ):
        button = dialog.get_widget_for_response(response_id)
        if button is not None:
            _apply_alt_font(button)

    content = dialog.get_content_area()
    content.set_spacing(8)

    # TRANSLATORS: Question shown in the font check dialog displayed
    # on startup on Windows. There are bugs that for some users the
    # font is blurry or resembles an alien language.
    text_quality_msg = _('Does all the text in this window look okay?')
    label_default = Gtk.Label(label=text_quality_msg)
    label_default.set_xalign(0.0)
    label_default.set_line_wrap(True)

    label_alt = Gtk.Label(label=text_quality_msg)
    label_alt.set_xalign(0.0)
    label_alt.set_line_wrap(True)
    _apply_alt_font(label_alt)

    content.pack_start(label_default, False, False, 0)
    content.pack_start(label_alt, False, False, 0)

    dialog.show_all()
    return dialog



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

import logging
import os
import stat
import sys
from importlib import import_module

import bleachbit
from bleachbit import IS_POSIX, IS_WINDOWS
from bleachbit.Language import get_text as _
from bleachbit.Network import unset_sslkeylogfile
from bleachbit.Options import options

if os.name == 'nt':
    from bleachbit import Windows

logger = logging.getLogger(__name__)


def _is_version_upgrade(old_version, target_version):
    """Check if upgrading from old_version to >= target_version"""
    try:
        old_parts = [int(x) for x in old_version.split('.')]
        target_parts = [int(x) for x in target_version.split('.')]

        # Pad shorter version with zeros
        max_len = max(len(old_parts), len(target_parts))
        old_parts.extend([0] * (max_len - len(old_parts)))
        target_parts.extend([0] * (max_len - len(target_parts)))

        # Check if old_version < target_version
        return old_parts < target_parts
    except (ValueError, AttributeError):
        return False


def _get_missing_dependencies():
    """Check for missing optional dependencies.

    Returns: list of missing dependency names
    """
    deps = ['chardet', 'psutil', 'requests', 'urllib3']
    if os.name == 'nt':
        deps.append('plyer')

    missing = []
    for dep in deps:
        try:
            import_module(dep)
        except ImportError:
            missing.append(dep)
    return sorted(missing)


def _get_posix_permission_issues(fstat, options_file):
    """Check POSIX-specific ownership and access issues.

    Returns (has_error, lines).
    """
    lines = []
    has_error = False
    import pwd  # pylint: disable=import-outside-toplevel
    from bleachbit.General import get_real_uid, get_real_username
    try:
        file_owner = pwd.getpwuid(fstat.st_uid).pw_name
    except (KeyError, OverflowError):
        file_owner = str(fstat.st_uid)
    current_uid = get_real_uid()
    try:
        current_user = get_real_username()
    except RuntimeError:
        has_error = True
        current_user = str(current_uid)
    lines.append(f'File owner: {file_owner} (uid {fstat.st_uid})')
    lines.append(f'Current user: {current_user} (uid {current_uid})')
    if fstat.st_uid != current_uid:
        has_error = True
        lines.append('File owner does not match current user')
    # os.access checks
    lines.append(f'os.access R_OK: {os.access(options_file, os.R_OK)}')
    lines.append(f'os.access W_OK: {os.access(options_file, os.W_OK)}')
    return has_error, lines


def _get_windows_user_info():
    """Get Windows user information.

    Returns (current_sid, current_name, token_groups).
    """
    if not IS_WINDOWS:
        raise RuntimeError("This function is only available on Windows")
    import win32api  # pylint: disable=import-outside-toplevel
    import win32file  # pylint: disable=import-outside-toplevel
    import win32security  # pylint: disable=import-outside-toplevel
    process_token = win32security.OpenProcessToken(
        win32api.GetCurrentProcess(),
        win32security.TOKEN_QUERY)
    current_sid = win32security.GetTokenInformation(
        process_token, win32security.TokenUser)[0]
    current_name = win32security.LookupAccountSid(
        None, current_sid)[0]
    token_groups = win32security.GetTokenInformation(
        process_token, win32security.TokenGroups)
    win32file.CloseHandle(process_token)
    # Convert PySID objects to string representation for hashing
    group_sids = {win32security.ConvertSidToStringSid(
        g[0]) for g in token_groups}
    current_sid_str = win32security.ConvertSidToStringSid(current_sid)
    return current_sid_str, current_name, group_sids


def _get_windows_file_owner(filepath):
    """Get the owner of a file on Windows.

    Returns (owner_sid_str, owner_name).
    """
    if not IS_WINDOWS:
        raise RuntimeError("This function is only available on Windows")
    import win32security  # pylint: disable=import-outside-toplevel
    file_sd = win32security.GetFileSecurity(
        filepath, win32security.OWNER_SECURITY_INFORMATION)
    file_owner_sid = file_sd.GetSecurityDescriptorOwner()
    file_owner_name = win32security.LookupAccountSid(
        None, file_owner_sid)[0]
    file_owner_sid_str = win32security.ConvertSidToStringSid(file_owner_sid)
    return file_owner_sid_str, file_owner_name


def _get_windows_permission_issues(options_file):
    """Check Windows-specific ownership issues.

    Returns (has_error, lines).
    """
    if not IS_WINDOWS:
        raise RuntimeError("This function is only available on Windows")
    lines = []
    has_error = False
    try:
        file_owner_sid_str, file_owner_name = _get_windows_file_owner(
            options_file)
        current_sid, current_name, group_sids = _get_windows_user_info()
        lines.append(f'File owner: {file_owner_name}')
        lines.append(f'Current user: {current_name}')
        # Administrators owns the directory, and the created file inherits
        # from the owner. This is not representative of a normal installation.
        if 'APPVEYOR' in os.environ and file_owner_name == 'Administrators':
            lines.append('Skipping file ownership check on AppVeyor')
            return False, lines
        if file_owner_sid_str != current_sid and file_owner_sid_str not in group_sids:
            has_error = True
            lines.append('File owner does not match current user')
    except Exception as e:
        has_error = True
        lines.append(f'Could not check file ownership: {e}')
    return has_error, lines


def _get_config_permission_issues():
    """Check for permission and file access issues with the
    configuration.

    If no errors, returns False.
    If errors found, returns list of strings with info.
    """
    lines = []
    has_error = False

    _options_file = bleachbit.options_file
    _options_dir = bleachbit.options_dir

    try:
        with open(_options_file, 'rb') as f:
            f.read(1)
    except (IOError, OSError, PermissionError) as e:
        has_error = True
        lines.append(f"Read error: {type(e).__name__}: {e}")

    try:
        with open(_options_file, 'a') as f:
            f.write('')
    except (IOError, OSError, PermissionError) as e:
        has_error = True
        lines.append(f"Write error: {type(e).__name__}: {e}")

    try:
        fstat = os.stat(_options_file)
        mode = stat.filemode(fstat.st_mode)
        lines.append(f'File: {_options_file}')
        lines.append(
            f'Permissions: {mode} ({oct(stat.S_IMODE(fstat.st_mode))})')
    except OSError as e:
        has_error = True
        lines.append(f'Cannot stat {_options_file}: {e}')

    # Directory info
    try:
        dstat = os.stat(_options_dir)
        dmode = stat.filemode(dstat.st_mode)
        lines.append(f'Directory: {_options_dir}')
        lines.append(
            f'Directory permissions: {dmode} ({oct(stat.S_IMODE(dstat.st_mode))})')
    except OSError as e:
        has_error = True
        lines.append(f'Cannot stat directory {_options_dir}: {e}')

    # File ownership vs current user
    if 'fstat' in locals():
        if IS_POSIX:
            _has_error, _lines = _get_posix_permission_issues(
                fstat, _options_file)
            has_error = has_error or _has_error
            lines.extend(_lines)
        elif IS_WINDOWS:
            _has_error, _lines = _get_windows_permission_issues(
                _options_file)
            has_error = has_error or _has_error
            lines.extend(_lines)

    if has_error:
        return lines
    return False


def _has_selected_warning_option():
    """Check if any selected cleaning option has a warning.

    Returns True if any selected option has a warning, False otherwise.
    """
    from bleachbit.Cleaner import backends  # pylint: disable=import-outside-toplevel
    if not backends:
        return False
    for cleaner_id, cleaner in backends.items():
        for (o_id, _o_name) in cleaner.get_options():
            if options.get_tree(cleaner_id, o_id):
                if cleaner.get_warning(o_id):
                    return True
    return False


def get_startup_messages(auto_exit):
    """Return a list of startup messages

    Can have a side effects
        - set first_start to False
        - unset environment variable

    Return: list of tuples (message, is_error) where is_error is True
    for error messages.
    """
    ret_msgs = []

    perm_issues_list = _get_config_permission_issues()
    if perm_issues_list:
        perm_issues_str = '\n'.join(perm_issues_list)
        ret_msgs.append((
            f'The configuration file {bleachbit.options_file} has a permissions issue, so existing '
            'preferences might not be loaded, and changes may not be saved.\n\n'
            f'{perm_issues_str}', True))

    missing_deps = _get_missing_dependencies()

    if missing_deps:
        ret_msgs.append((
            f'Missing optional Python packages: {", ".join(missing_deps)}. '
            'Some features may be limited.', True))

    # Call unset_sslkeylogfile before checking for updates.
    if unset_sslkeylogfile(False):
        ret_msgs.append((
            'The environment variable SSLKEYLOGFILE is not supported', True))

    # Check if any selected cleaning option has a warning
    if not options.get('expert_mode') and _has_selected_warning_option():
        ret_msgs.append((
            # TRANSLATORS: "Expert Mode" is the name of a feature/setting found in
            # Preferences. Keep it consistent with the translation of "Expert Mode"
            # in the preferences dialog. This message appears when cleaning options
            # with warnings are selected but expert mode is disabled.
            _('One or more selected cleaning options have warnings and '
              'require Expert Mode to clean. Enable it in preferences.'), False))

    if not options.get('delete_confirmation') and not options.get('expert_mode'):
        ret_msgs.append((
            # TRANSLATORS: "Expert Mode" is the name of a feature/setting found in
            # Preferences. Keep it consistent with the translation of "Expert Mode"
            # in the preferences dialog. This message appears when the user tries to
            # skip a delete confirmation without enabling Expert Mode.
            _('Delete confirmation is disabled but expert mode is not enabled. '
              'Enable Expert Mode in preferences to skip delete confirmations.'), False))

    # Show information for first start.
    # (The first start flag is set also for each new version.)
    if options.get("first_start") and not auto_exit:
        if os.name == 'posix':
            ret_msgs.append((
                # TRANSLATORS: First-start hint for Linux users shown in
                # the log on the main screen.
                _('Access the application menu by clicking the hamburger icon on the title bar.'), False))
        elif os.name == 'nt':
            ret_msgs.append((
                # TRANSLATORS: First-start hint for Windows users shown in
                # the log on the main screen.
                _('Access the application menu by clicking the logo on the title bar.'), False))
        options.set('first_start', False)

    # Show notice about admin privileges.
    if os.name == 'posix' and os.path.expanduser('~') == '/root':
        ret_msgs.append((
            # TRANSLATORS: Warning shown on startup when running BleachBit as root on Linux.
            # It means, for example, that cleaning a browser will clean root's browser data.
            _('You are running BleachBit with administrative privileges for cleaning '
              'shared parts of the system, and references to the user profile folder '
              'will clean only the root account.'), False))
    if os.name == 'nt':
        from win32com.shell.shell import IsUserAnAdmin
        if options.get('shred') and not IsUserAnAdmin():
            ret_msgs.append((
                # TRANSLATORS: Warning shown on startup on Windows.
                _('Run BleachBit with administrator privileges to improve the '
                  'accuracy of overwriting the contents of files.'), False))
        if Windows.is_ots_elevation():
            ret_msgs.append((
                # TRANSLATORS: Warning shown on startup on Windows when elevated
                # with different account. It means, for example, that cleaning a
                # browser will clean the browser data of the administrator account.
                _('You elevated privileges using a different account to clean '
                  'shared parts of the system. User-specific paths will refer to '
                  'the administrator account, so to clean your profile, run '
                  'BleachBit again as a standard user.'), False))

        if 'windowsapps' in sys.executable.lower():
            ret_msgs.append((
                # TRANSLATORS: Warning shown on startup on Windows when running
                # unofficial Microsoft Store version. Advises user to get genuine
                # version from official website.
                _('There is no official version of BleachBit on the Microsoft Store. '
                  'Get the genuine version at https://www.bleachbit.org where it is '
                  'always free of charge.'), False))

    return ret_msgs



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.


"""Constants"""

from bleachbit.Language import get_text as _

# This module centralizes constants.
# By reducing imports, this module avoids circular dependencies.

URL_COOKIE_MGR = "https://docs.bleachbit.org/doc/cookie-manager.html"

# TRANSLATORS: Button label on (1) the headerbar and (2) the chaff dialog.
# 'Abort' is a verb.
ABORT_BUTTON_LABEL = _('Abort')

# This string is used several times. Listing it here helps with translation.
# TRANSLATORS: This is the name of a cleaning action. 'Clean' is a verb.
# It shows in the log of actions that would be performed (preview mode) or
# in the log of actions that are performed (clean mode).
CLEAN_FILE_LABEL = _('Clean file')

# TRANSLATORS: This message is used in three places (1) label at top of the
# preferences dialog (2) confirmation dialog when enabling the cleaning option
# (3) log message when starting to clean empty space.
EMPTY_SPACE_WARNING = _("Wiping empty space removes traces of files that were "
                        "deleted without shredding, but it will not free additional disk space. The "
                        "process can take a very long time and may temporarily slow your computer. "
                        "It is not necessary if your drive is protected with full-disk encryption. "
                        "The method works best on traditional hard drives. On solid-state drives, it "
                        "is less reliable, and frequent use contributes to wear.")

# TRANSLATORS: Error message shown in the infobar or notice shown in a tooltip.
# Expert mode is an option in the preferences to relax guardrails.
REQUIRES_EXPERT_MODE = _("This option requires expert mode. "
                         "Enable it in Preferences.")



# vim: ts=4:sw=4:expandtab

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
GTK graphical user interface
"""

# standard library
import logging

from bleachbit.GtkShim import require_gtk
from bleachbit.Log import set_root_log_level
from bleachbit.Options import options  # keep

# Ensure GTK is available for this GUI module
require_gtk()

# Now that the configuration is loaded, honor the debug preference there.
set_root_log_level(options.get('debug'))
logger = logging.getLogger(__name__)



# vim: ts=4:sw=4:expandtab
# -*- coding: UTF-8 -*-

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Wipe memory
"""

import logging
import os
import re
import subprocess
import sys

from bleachbit import FileUtilities
from bleachbit import General
from bleachbit.Language import get_text as _
from bleachbit.Wipe import wipe_contents

logger = logging.getLogger(__name__)


def count_swap_linux():
    """Count the number of swap devices in use"""
    count = 0
    with open("/proc/swaps") as f:
        for line in f:
            if line[0] == '/':
                count += 1
    return count


def get_proc_swaps():
    """Return the output of 'swapon -s'"""
    # Usually 'swapon -s' is identical to '/proc/swaps'
    # Here is one exception:
    # https://bugs.launchpad.net/ubuntu/+source/bleachbit/+bug/1092792
    (rc, stdout, _stderr) = General.run_external(['swapon', '-s'])
    if 0 == rc:
        return stdout
    logger.debug(
        _("The command 'swapoff -s' failed, so falling back to /proc/swaps for swap information."))
    return open("/proc/swaps").read()


def parse_swapoff(swapoff):
    """Parse the output of swapoff and return the device name"""
    # English is 'swapoff on /dev/sda5' but German is 'swapoff für ...'
    # Example output in English with LVM and hyphen: 'swapoff on /dev/mapper/lubuntu-swap_1'
    # This matches swap devices and swap files
    ret = re.search(r'^swapoff (\w* )?(/[\w/.-]+)$', swapoff)
    if not ret:
        # no matches
        return None
    return ret.group(2)


def disable_swap_linux():
    """Disable Linux swap and return list of devices"""
    if 0 == count_swap_linux():
        return
    logger.debug(_("Disabling swap."))
    args = ["swapoff", "-a", "-v"]
    (rc, stdout, stderr) = General.run_external(args)
    if 0 != rc:
        raise RuntimeError(stderr.replace("\n", ""))
    devices = []
    for line in stdout.split('\n'):
        line = line.replace('\n', '')
        if '' == line:
            continue
        ret = parse_swapoff(line)
        if ret is None:
            raise RuntimeError(f"Unexpected output:\nargs='{args}'\n"
                               f"stdout='{stdout}'\nstderr='{stderr}'")
        devices.append(ret)
    return devices


def enable_swap_linux():
    """Enable Linux swap"""
    logger.debug(_("Re-enabling swap."))
    args = ["swapon", "-a"]
    p = subprocess.Popen(args, stderr=subprocess.PIPE)
    p.wait()
    outputs = p.communicate()
    if 0 != p.returncode:
        raise RuntimeError(outputs[1].replace("\n", ""))


def make_self_oom_target_linux():
    """Make the current process the primary target for Linux out-of-memory killer"""
    path = f'/proc/{os.getpid()}/oom_score_adj'
    if os.path.exists(path):
        with open(path, 'w', encoding='utf-8') as f:
            f.write('1000')
    # OOM likes nice processes
    logger.debug(_("Setting nice value %d for this process."), os.nice(19))
    # OOM prefers non-privileged processes
    try:
        uid = General.get_real_uid()
        if uid > 0:
            # TRANSLATORS: Debug message when a process gives up root/admin privileges.
            # %(pid)d is the integer process ID; %(uid)d is the integer user ID to switch to.
            drop_msg = _("Dropping privileges of process ID %(pid)d to user ID %(uid)d.")
            logger.debug(drop_msg, {'pid': os.getpid(), 'uid': uid})
            os.seteuid(uid)
    except:
        logger.exception('Error when dropping privileges')


def fill_memory_linux():
    """Fill unallocated memory"""
    report_free()
    allocbytes = int(physical_free() * 0.4)
    if allocbytes < 1024:
        return
    bytes_str = FileUtilities.bytes_to_human(allocbytes)
    # TRANSLATORS: The variable is a quantity like 5kB
    logger.info(_("Allocating and wiping %s of memory."),
                bytes_str)
    try:
        buf = '\x00' * allocbytes
    except MemoryError:
        pass
    else:
        fill_memory_linux()
        # TRANSLATORS: The variable is a quantity like 5kB
        logger.debug(_("Freeing %s of memory."), bytes_str)
        del buf
    report_free()


def get_swap_size_linux(device, proc_swaps=None):
    """Return the size of the partition in bytes"""
    if proc_swaps is None:
        proc_swaps = get_proc_swaps()
    line = proc_swaps.split('\n')[0]
    if not re.search(r'Filename\s+Type\s+Size', line):
        raise RuntimeError("Unexpected first line in swap summary '%s'" % line)
    for line in proc_swaps.split('\n')[1:]:
        ret = re.search(r"%s\s+\w+\s+([0-9]+)\s" % device, line)
        if ret:
            return int(ret.group(1)) * 1024
    raise RuntimeError("error: cannot find size of swap device '%s'\n%s" %
                       (device, proc_swaps))


def get_swap_uuid(device):
    """Find the UUID for the swap device"""
    uuid = None
    args = ['blkid', device, '-s', 'UUID']
    (_rc, stdout, _stderr) = General.run_external(args)
    for line in stdout.split('\n'):
        # example: /dev/sda5: UUID="ee0e85f6-6e5c-42b9-902f-776531938bbf"
        ret = re.search(r"^%s: UUID=\"([a-z0-9-]+)\"" % device, line)
        if ret is not None:
            uuid = ret.group(1)
    # TRANSLATORS: Debug message. 'Found' is a past tense verb (short for
    # "Found [that] the UUID for swap device ..."). %(device)s is the device
    # path (e.g., /dev/sda5); %(uuid)s is a UUID string
    # (e.g., ee0e85f6-6e5c-42b9-902f-776531938bbf). Do not translate variables.
    logger.debug(_("Found UUID for swap device %(device)s is %(uuid)s."),
                 {'device': device, 'uuid': uuid})
    return uuid


def physical_free_darwin(run_vmstat=None):
    def parse_line(k, v):
        return k, int(v.strip(" ."))

    def get_page_size(line):
        m = re.match(
            r"Mach Virtual Memory Statistics: \(page size of (\d+) bytes\)",
            line)
        if m is None:
            raise RuntimeError("Can't parse vm_stat output")
        return int(m.groups()[0])
    if run_vmstat is None:
        def run_vmstat():
            return subprocess.check_output(["vm_stat"])
    output = iter(run_vmstat().split("\n"))
    page_size = get_page_size(next(output))
    vm_stat = dict(parse_line(*l.split(":")) for l in output if l != "")
    return vm_stat["Pages free"] * page_size


def physical_free_linux():
    """Return the physical free memory on Linux"""
    free_bytes = 0
    with open("/proc/meminfo") as f:
        for line in f:
            line = line.replace("\n", "")
            ret = re.search(r'(MemFree|Cached):[ ]*([0-9]*) kB', line)
            if ret is not None:
                kb = int(ret.group(2))
                free_bytes += kb * 1024
    if free_bytes > 0:
        return free_bytes
    else:
        raise Exception("unknown")


def physical_free_windows():
    """Return physical free memory on Windows"""

    from ctypes import c_ulong, c_ulonglong, Structure, sizeof, windll, byref

    class MEMORYSTATUSEX(Structure):
        _fields_ = [
            ('dwLength', c_ulong),
            ('dwMemoryLoad', c_ulong),
            ('ullTotalPhys', c_ulonglong),
            ('ullAvailPhys', c_ulonglong),
            ('ullTotalPageFile', c_ulonglong),
            ('ullAvailPageFile', c_ulonglong),
            ('ullTotalVirtual', c_ulonglong),
            ('ullAvailVirtual', c_ulonglong),
            ('ullExtendedVirtual', c_ulonglong),
        ]

    def GlobalMemoryStatusEx():
        x = MEMORYSTATUSEX()
        x.dwLength = sizeof(x)
        windll.kernel32.GlobalMemoryStatusEx(byref(x))
        return x

    z = GlobalMemoryStatusEx()
    return z.ullAvailPhys


def physical_free():
    if sys.platform == 'linux':
        return physical_free_linux()
    elif 'win32' == sys.platform:
        return physical_free_windows()
    elif 'darwin' == sys.platform:
        return physical_free_darwin()
    else:
        raise RuntimeError('unsupported platform for physical_free()')


def report_free():
    """Report free memory"""
    bytes_free = physical_free()
    bytes_str = FileUtilities.bytes_to_human(bytes_free)
    # TRANSLATORS: The variable is a quantity like 5kB
    logger.debug(_("Physical free memory is %s."),
                 bytes_str)


def wipe_swap_linux(devices, proc_swaps):
    """Shred the Linux swap file and then reinitialize it"""
    if devices is None:
        return
    if 0 < count_swap_linux():
        raise RuntimeError('Cannot wipe swap while it is in use')
    for device in devices:
        # if '/cryptswap' in device:
        #    logger.info('Skipping encrypted swap device %s.', device)
        #    continue
        # TRANSLATORS: The variable is a device like /dev/sda2
        logger.info(_("Wiping the swap device %s."), device)
        safety_limit_bytes = 29 * 1024 ** 3  # 29 gibibytes
        actual_size_bytes = get_swap_size_linux(device, proc_swaps)
        if actual_size_bytes > safety_limit_bytes:
            raise RuntimeError(
                f'swap device {device} is larger ({actual_size_bytes})'
                f' than expected ({safety_limit_bytes})')
        uuid = get_swap_uuid(device)
        # wipe
        wipe_contents(device, truncate=False)
        # reinitialize
        # TRANSLATORS: The variable is a device like /dev/sda2
        logger.debug(_("Reinitializing the swap device %s."), device)
        args = ['mkswap', device]
        if uuid:
            args.append("-U")
            args.append(uuid)
        (rc, _stdout, stderr) = General.run_external(args)
        if 0 != rc:
            raise RuntimeError(stderr.replace("\n", ""))


def wipe_memory():
    """Wipe unallocated memory"""
    for cmd in ('swapon', 'swapoff', 'blkid'):
        if not FileUtilities.exe_exists(cmd):
            raise RuntimeError(f"wipe_memory: Command {cmd} not found")
    # cache the file because 'swapoff' changes it
    proc_swaps = get_proc_swaps()
    devices = disable_swap_linux()
    yield True  # process GTK+ idle loop
    # TRANSLATORS: The variable is a device like /dev/sda2
    logger.debug(_("Detected these swap devices: %s"), str(devices))
    wipe_swap_linux(devices, proc_swaps)
    yield True
    child_pid = os.fork()
    if 0 == child_pid:
        make_self_oom_target_linux()
        fill_memory_linux()
        os._exit(0)
    else:
        # TRANSLATORS: This is a debugging message that the parent process
        # is waiting for the child process. %(parent_pid)d is the parent
        # process ID; %(child_pid)d is the child process ID.
        logger.debug(_("The function wipe_memory() with process ID %(parent_pid)d is "
                       "waiting for child process ID %(child_pid)d."),
                     {'parent_pid': os.getpid(), 'child_pid': child_pid})
        rc = os.waitpid(child_pid, 0)[1]
        if rc not in [0, 9]:
            logger.warning(
                _("The child memory-wiping process returned code %d."), rc)
    enable_swap_linux()
    yield 0  # how much disk space was recovered
